from typing import Union

from packaging import version as packaging_version

from sickchill import settings, version
from sickchill.init_helpers import pip_install
from sickchill.oldbeard import helpers, notifiers

from .abstract import UpdateManagerBase


class PipUpdateManager(UpdateManagerBase):
    def __init__(self):
        self._newest_version: Union[packaging_version.LegacyVersion, packaging_version.Version] = None
        self.session = helpers.make_session()
        self.branch = "pip"

    def get_current_version(self) -> Union[packaging_version.LegacyVersion, packaging_version.Version]:
        return packaging_version.parse(version.__version__)

    def get_clean_version(self, use_version: packaging_version.Version = None):
        _version = use_version or self.get_current_version()
        result = f"v{_version.major:04d}.{_version.minor:02d}.{_version.micro:02}"
        if _version.is_postrelease:
            result += f"-{_version.post}"

        return result

    def get_current_commit_hash(self) -> Union[packaging_version.LegacyVersion, packaging_version.Version]:
        return self.get_current_version()

    def get_newest_version(self) -> Union[packaging_version.LegacyVersion, packaging_version.Version]:
        self._newest_version = packaging_version.parse(self.session.get("https://pypi.org/pypi/sickchill/json").json()["info"]["version"])
        return self._newest_version

    def get_newest_commit_hash(self) -> Union[packaging_version.LegacyVersion, packaging_version.Version]:
        return self._newest_version

    def get_num_commits_behind(self):
        if not self.need_update():
            return 0

        newest, current = self._newest_version, self.get_current_version()
        return f"Major: {newest.major - current.major}, Minor: {newest.minor - current.minor}, Micro: {newest.micro - current.micro}"

    def list_remote_branches(self):
        return ["pip"]

    def set_newest_text(self):
        if self.need_update():
            if self.get_newest_commit_hash():
                current = self.get_clean_version()
                newest = self.get_clean_version(self.get_newest_commit_hash())
                url = f"https://github.com/{settings.GIT_ORG}/{settings.GIT_REPO}/compare/{current}...{newest}"
            else:
                url = f"https://github.com/{settings.GIT_ORG}/{settings.GIT_REPO}/commits/"

            newest_tag = "newer_version_available"
            update_url = self.get_update_url()
            newest_text = _(
                f'There is a <a href="{url}" onclick="window.open(this.href); return false;">newer version available</a> &mdash; <a href="{update_url}">Update Now</a>'
            )

        else:
            return

        helpers.add_site_message(newest_text, tag=newest_tag, level="success")

    def need_update(self):
        return self.get_newest_version() > self.get_current_version()

    # def get_update_url(self):
    #     return self.session.get("https://pypi.org/pypi/sickchill/json").json()["info"]["release_url"]

    def update(self):
        if not pip_install("sickchill"):
            return False

        settings.CUR_COMMIT_HASH = self._newest_version
        notifiers.notify_git_update(settings.CUR_COMMIT_HASH or "")
        return True



import os

from tornado.web import addslash

from sickchill import settings
from sickchill.update_manager import UpdateManager
from sickchill.views.common import PageTemplate
from sickchill.views.index import WebRoot
from sickchill.views.routes import Route


@Route("/config(/?.*)", name="config:main")
class Config(WebRoot):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @staticmethod
    def ConfigMenu():
        menu = [
            {"title": _("General"), "path": "config/general/", "icon": "fa fa-cog"},
            {"title": _("Backup/Restore"), "path": "config/backuprestore/", "icon": "fa fa-floppy-o"},
            {"title": _("Search Settings"), "path": "config/search/", "icon": "fa fa-search"},
            {"title": _("Search Providers"), "path": "config/providers/", "icon": "fa fa-plug"},
            {"title": _("Subtitles Settings"), "path": "config/subtitles/", "icon": "fa fa-language"},
            {"title": _("Post Processing"), "path": "config/postProcessing/", "icon": "fa fa-refresh"},
            {"title": _("Notifications"), "path": "config/notifications/", "icon": "fa fa-bell-o"},
            {"title": _("Anime"), "path": "config/anime/", "icon": "fa fa-eye"},
        ]

        return menu

    @addslash
    def index(self):
        t = PageTemplate(rh=self, filename="config.mako")
        try:
            import getpass

            sc_user = getpass.getuser()
        except Exception:
            sc_user = os.getuid()

        try:
            import locale

            sc_locale = locale.getdefaultlocale()
        except Exception:
            sc_locale = "Unknown", "Unknown"

        try:
            import ssl

            ssl_version = ssl.OPENSSL_VERSION
        except Exception:
            ssl_version = "Unknown"

        sc_version = ""
        if settings.VERSION_NOTIFY or settings.BRANCH == "pip":
            updater = UpdateManager()
            if updater:
                if settings.BRANCH == "pip":
                    sc_version = updater.updater.get_clean_version()
                else:
                    updater.need_update()
                    sc_version = updater.get_current_version()

        return t.render(
            submenu=self.ConfigMenu(),
            title=_("SickChill Configuration"),
            header=_("SickChill Configuration"),
            topmenu="config",
            sc_user=sc_user,
            sc_locale=sc_locale,
            ssl_version=ssl_version,
            sc_version=sc_version,
        )



import os
import string

from sickchill import logger, settings


# adapted from http://stackoverflow.com/questions/827371/is-there-a-way-to-list-all-the-available-drive-letters-in-python/827490
def getWinDrives():
    """ Return list of detected drives """
    assert os.name == "nt"
    from ctypes import windll

    drives = []
    bitmask = windll.kernel32.GetLogicalDrives()  # @UndefinedVariable
    for letter in string.ascii_uppercase:
        if bitmask & 1:
            drives.append(letter)
        bitmask >>= 1

    return drives


def getFileList(path, includeFiles, fileTypes):
    # prune out directories to protect the user from doing stupid things (already lower case the dir to reduce calls)
    hide_list = [
        "boot",
        "bootmgr",
        "cache",
        "config.msi",
        "msocache",
        "recovery",
        "$recycle.bin",
        "recycler",
        "system volume information",
        "temporary internet files",
    ]  # windows specific
    hide_list += [".fseventd", ".spotlight", ".trashes", ".vol", "cachedmessages", "caches", "trash"]  # osx specific
    hide_list += [".git"]

    file_list, dir_list = [], []
    for filename in os.listdir(path):
        if filename.lower() in hide_list:
            continue

        full_filename = os.path.join(path, filename)
        is_file = os.path.isfile(full_filename)

        if not includeFiles and is_file:
            continue

        is_image = False
        allowed_type = True
        if is_file and fileTypes:
            if "images" in fileTypes:
                is_image = filename.endswith(("jpg", "jpeg", "png", "tiff", "gif"))
            allowed_type = filename.endswith(tuple(fileTypes)) or is_image

            if not allowed_type:
                continue

        item_to_add = {"name": filename, "path": full_filename, "isFile": is_file, "isImage": is_image, "isAllowed": allowed_type}

        if is_file:
            file_list.append(item_to_add)
        else:
            dir_list.append(item_to_add)

    # Sort folders first, alphabetically, case insensitive
    dir_list.sort(key=lambda mbr: mbr.get("name").lower())
    file_list.sort(key=lambda mbr: mbr.get("name").lower())
    return dir_list + file_list


def foldersAtPath(path, includeParent=False, includeFiles=False, fileTypes=None):
    """
    Returns a list of dictionaries with the folders contained at the given path.

    Give the empty string as the path to list the contents of the root path
    (under Unix this means "/", on Windows this will be a list of drive letters)

    :param path: to list contents
    :param includeParent: boolean, include parent dir in list as well
    :param includeFiles: boolean, include files or only directories
    :param fileTypes: list, file extensions to include, 'images' is an alias for image types
    :return: list of folders/files
    """

    # walk up the tree until we find a valid directory path
    while path and not os.path.isdir(path) and path != "/":
        if path == os.path.dirname(path):
            path = ""
        else:
            path = os.path.dirname(path)

    if path == "":
        if os.name != "nt":
            path = "/"
        else:
            entries = [{"currentPath": "Root"}]
            for letter in getWinDrives():
                letter_path = letter + ":\\"
                entries.append({"name": letter_path, "path": letter_path})

            for name, share in settings.WINDOWS_SHARES.items():
                entries.append({"name": name, "path": r"\\{server}\{path}".format(server=share["server"], path=share["path"])})

            return entries

    # fix up the path and find the parent
    path = os.path.abspath(os.path.normpath(path))
    parent_path = os.path.dirname(path)

    # if we're at the root then the next step is the meta-node showing our drive letters
    if path == parent_path and os.name == "nt":
        parent_path = ""

    fileTypes = fileTypes or []

    try:
        file_list = getFileList(path, includeFiles, fileTypes)
    except OSError as e:
        logger.warning("Unable to open {0}: {1} / {2}".format(path, repr(e), str(e)))
        file_list = getFileList(parent_path, includeFiles, fileTypes)

    entries = [{"currentPath": path}]
    if path == "/":
        for name, share in settings.WINDOWS_SHARES.items():
            entries.append({"name": name, "path": r"\\{server}\{path}".format(server=share["server"], path=share["path"])})

    if includeParent and parent_path != path:
        entries.append({"name": "..", "path": parent_path})
    entries.extend(file_list)

    return entries



from sickchill.oldbeard.common import Quality, qualityPresetStrings


def get_quality_string(quality):
    """
    :param quality: The quality to convert into a string
    :return: The string representation of the provided quality
    """

    if quality in qualityPresetStrings:
        return qualityPresetStrings[quality]

    if quality in Quality.qualityStrings:
        return Quality.qualityStrings[quality]

    return "Custom"



import json
from urllib.parse import urlencode

from sickchill import logger
from sickchill.helper.common import convert_size, try_int
from sickchill.helper.exceptions import AuthException
from sickchill.oldbeard import tvcache
from sickchill.providers.torrent.TorrentProvider import TorrentProvider


class Provider(TorrentProvider):
    """Main provider object"""

    def __init__(self):
        """ Initialize the class """
        super().__init__("Norbits")

        self.username = None
        self.passkey = None
        self.minseed = 0
        self.minleech = 0

        self.cache = tvcache.TVCache(self, min_time=20)  # only poll Norbits every 15 minutes max

        self.url = "https://norbits.net"
        self.urls = {"search": self.url + "/api2.php?action=torrents", "download": self.url + "/download.php?"}

    def _check_auth(self):

        if not self.username or not self.passkey:
            raise AuthException(("Your authentication credentials for {} are " "missing, check your config.").format(self.name))

        return True

    @staticmethod
    def _check_auth_from_data(parsed_json):
        """ Check that we are authenticated. """

        if "status" in parsed_json and "message" in parsed_json and parsed_json.get("status") == 3:
            logger.warning("Invalid username or password. Check your settings")

        return True

    def search(self, search_params, age=0, ep_obj=None):
        """ Do the actual searching and JSON parsing"""

        results = []

        for mode in search_params:
            items = []
            logger.debug(_("Search Mode: {mode}".format(mode=mode)))

            for search_string in search_params[mode]:
                if mode != "RSS":
                    logger.debug(_("Search String: {search_string}".format(search_string=search_string)))

                post_data = {
                    "username": self.username,
                    "passkey": self.passkey,
                    "category": "2",  # TV Category
                    "search": search_string,
                }

                self._check_auth()
                parsed_json = self.get_url(self.urls["search"], post_data=json.dumps(post_data), returns="json")

                if not parsed_json:
                    return results

                if self._check_auth_from_data(parsed_json):
                    json_items = parsed_json.get("data", "")
                    if not json_items:
                        logger.exception("Resulting JSON from provider is not correct, " "not parsing it")

                    for item in json_items.get("torrents", []):
                        title = item.pop("name", "")
                        download_url = "{0}{1}".format(self.urls["download"], urlencode({"id": item.pop("id", ""), "passkey": self.passkey}))

                        if not all([title, download_url]):
                            continue

                        seeders = try_int(item.pop("seeders", 0))
                        leechers = try_int(item.pop("leechers", 0))

                        if seeders < self.minseed or leechers < self.minleech:
                            logger.debug(
                                "Discarding torrent because it does not meet "
                                "the minimum seeders or leechers: {0} (S:{1} L:{2})".format(title, seeders, leechers)
                            )
                            continue

                        info_hash = item.pop("info_hash", "")
                        size = convert_size(item.pop("size", -1), -1)

                        item = {"title": title, "link": download_url, "size": size, "seeders": seeders, "leechers": leechers, "hash": info_hash}
                        if mode != "RSS":
                            logger.debug(
                                _(
                                    "Found result: {title} with {seeders} seeders and {leechers} leechers".format(
                                        title=title, seeders=seeders, leechers=leechers
                                    )
                                )
                            )

                        items.append(item)
            # For each search mode sort all the items by seeders if available
            items.sort(key=lambda d: try_int(d.get("seeders", 0)), reverse=True)

            results += items

        return results



from datetime import date, timedelta
from operator import itemgetter

from sickchill import settings
from sickchill.helper.common import dateFormat, timeFormat
from sickchill.helper.quality import get_quality_string
from sickchill.oldbeard.common import Quality, UNAIRED, WANTED
from sickchill.oldbeard.db import DBConnection
from sickchill.oldbeard.network_timezones import parse_date_time
from sickchill.oldbeard.sbdatetime import sbdatetime

SNATCHED = Quality.SNATCHED + Quality.SNATCHED_PROPER + Quality.SNATCHED_BEST  # type = list


class ComingEpisodes(object):
    """
    Snatched: snatched but not yet processed (no re-downloads)
    Missed:   yesterday...(less than settings.COMING_EPS_MISSED_RANGE)
    Today:    today
    Soon:     tomorrow till next week
    Later:    later than next week
    """

    categories = ["snatched", "missed", "today", "soon", "later"]
    sorts = {
        "date": itemgetter("snatchedsort", "localtime", "episode"),
        "network": itemgetter("network", "localtime", "episode"),
        "show": itemgetter("show_name", "localtime", "episode"),
    }

    def __init__(self):
        pass

    @staticmethod
    def get_coming_episodes(categories, sort, group, paused=settings.COMING_EPS_DISPLAY_PAUSED):
        """
        :param categories: The categories of coming episodes. See ``ComingEpisodes.categories``
        :param sort: The sort to apply to the coming episodes. See ``ComingEpisodes.sorts``
        :param group: ``True`` to group the coming episodes by category, ``False`` otherwise
        :param paused: ``True`` to include paused shows, ``False`` otherwise
        :return: The list of coming episodes
        """

        categories = ComingEpisodes._get_categories(categories)
        sort = ComingEpisodes._get_sort(sort)

        today = date.today().toordinal()
        recently = (date.today() - timedelta(days=settings.COMING_EPS_MISSED_RANGE)).toordinal()
        next_week = (date.today() + timedelta(days=7)).toordinal()

        db = DBConnection(row_type="dict")
        fields_to_select = ", ".join(
            [
                "airdate",
                "airs",
                "e.description as description",
                "episode",
                "imdb_id",
                "e.indexer",
                "indexer_id",
                "e.location",
                "name",
                "network",
                "paused",
                "quality",
                "runtime",
                "season",
                "show_name",
                "showid",
                "e.status as epstatus",
                "s.status",
            ]
        )

        status_list = [WANTED, UNAIRED] + SNATCHED

        sql_l = []
        for show_obj in settings.showList:
            next_air_date = show_obj.nextEpisode()
            sql_l.append(
                [
                    "SELECT DISTINCT {0} ".format(fields_to_select) + "FROM tv_episodes e, tv_shows s "
                    "WHERE showid = ? "
                    "AND airdate <= ? "
                    "AND airdate >= ? "
                    "AND s.indexer_id = e.showid "
                    "AND e.status IN (" + ",".join(["?"] * len(status_list)) + ")",
                    [show_obj.indexerid, next_air_date or today, recently] + status_list,
                ]
            )

        results = []
        for sql_i in sql_l:
            if results:
                results += db.select(*sql_i)
            else:
                results = db.select(*sql_i)

        for index, item in enumerate(results):
            results[index]["localtime"] = sbdatetime.convert_to_setting(parse_date_time(item["airdate"], item["airs"], item["network"]))
            results[index]["snatchedsort"] = int(not results[index]["epstatus"] in SNATCHED)

        results.sort(key=ComingEpisodes.sorts[sort])

        if not group:
            return results

        grouped_results = ComingEpisodes._get_categories_map(categories)

        for result in results:
            if result["paused"] and not paused:
                continue

            result["airs"] = str(result["airs"]).replace("am", " AM").replace("pm", " PM").replace("  ", " ")
            result["airdate"] = result["localtime"].toordinal()

            if result["epstatus"] in SNATCHED:
                if result["location"]:
                    continue
                else:
                    category = "snatched"
            elif result["airdate"] < today:
                category = "missed"
            elif result["airdate"] >= next_week:
                category = "later"
            elif result["airdate"] == today:
                category = "today"
            else:
                category = "soon"

            if categories and category not in categories:
                continue

            if not result["network"]:
                result["network"] = ""

            result["quality"] = get_quality_string(result["quality"])
            result["airs"] = sbdatetime.sbftime(result["localtime"], t_preset=timeFormat).lstrip("0").replace(" 0", " ")
            result["weekday"] = 1 + result["localtime"].weekday()
            result["tvdbid"] = result["indexer_id"]
            result["airdate"] = sbdatetime.sbfdate(result["localtime"], d_preset=dateFormat)
            result["localtime"] = result["localtime"].toordinal()

            grouped_results[category].append(result)

        return grouped_results

    @staticmethod
    def _get_categories(categories):
        if not categories:
            return []

        if not isinstance(categories, list):
            return categories.split("|")

        return categories

    @staticmethod
    def _get_categories_map(categories):
        if not categories:
            return {}

        return {category: [] for category in categories}

    @staticmethod
    def _get_sort(sort):
        sort = sort.lower() if sort else ""

        if sort not in list(ComingEpisodes.sorts):
            return "date"

        return sort



import logging
import os
import shutil
import stat
import traceback

import validators
from rarfile import BadRarFile, Error, NeedFirstVolume, PasswordRequired, RarCRCError, RarExecError, RarFile, RarOpenError, RarWrongPassword

from sickchill import logger, settings
from sickchill.helper.common import is_sync_file, is_torrent_or_nzb_file
from sickchill.helper.exceptions import EpisodePostProcessingFailedException, FailedPostProcessingFailedException

from . import common, db, failedProcessor, helpers, postProcessor
from .name_parser.parser import InvalidNameException, InvalidShowException, NameParser


class ProcessResult(object):
    def __init__(self):
        self.result = True
        self.output = ""
        self.missed_files = []
        self.aggresult = True


def delete_folder(folder, check_empty=True):
    """
    Removes a folder from the filesystem

    :param folder: Path to folder to remove
    :param check_empty: Boolean, check if the folder is empty before removing it, defaults to True
    :return: True on success, False on failure
    """

    # check if it's a folder
    if not os.path.isdir(folder):
        return False

    # check if it isn't TV_DOWNLOAD_DIR
    if settings.TV_DOWNLOAD_DIR and helpers.real_path(folder) == helpers.real_path(settings.TV_DOWNLOAD_DIR):
        return False

    # check if it's empty folder when wanted checked
    if check_empty:
        check_files = os.listdir(folder)
        if check_files:
            logger.info("Not deleting folder {0} found the following files: {1}".format(folder, check_files))
            return False

        try:
            logger.info("Deleting folder (if it's empty): {0}".format(folder))
            os.rmdir(folder)
        except (OSError, IOError) as e:
            logger.warning("Warning: unable to delete folder: {0}: {1}".format(folder, str(e)))
            return False
    else:
        try:
            logger.info("Deleting folder: " + folder)
            shutil.rmtree(folder)
        except (OSError, IOError) as e:
            logger.warning("Warning: unable to delete folder: {0}: {1}".format(folder, str(e)))
            return False

    return True


def delete_files(process_path, unwanted_files, result, force=False):
    """
    Remove files from filesystem

    :param process_path: path to process
    :param unwanted_files: files we do not want
    :param result: Processor results
    :param force: Boolean, force deletion, defaults to false
    """
    if not result.result and force:
        result.output += log_helper("Forcing deletion of files, even though last result was not success", logger.DEBUG)
    elif not result.result:
        return

    # Delete all file not needed
    for cur_file in unwanted_files:
        cur_file_path = os.path.join(process_path, cur_file)
        if not os.path.isfile(cur_file_path):
            continue  # Prevent error when a notwantedfiles is an associated files

        result.output += log_helper("Deleting file: {0}".format(cur_file), logger.DEBUG)

        # check first the read-only attribute
        file_attribute = os.stat(cur_file_path)[0]
        if not file_attribute & stat.S_IWRITE:
            # File is read-only, so make it writeable
            result.output += log_helper("Changing ReadOnly Flag for file: {0}".format(cur_file), logger.DEBUG)
            try:
                os.chmod(cur_file_path, stat.S_IWRITE)
            except OSError as e:
                result.output += log_helper("Cannot change permissions of {0}: {1}".format(cur_file_path, str(e)), logger.DEBUG)
        try:
            os.remove(cur_file_path)
        except OSError as e:
            result.output += log_helper("Unable to delete file {0}: {1}".format(cur_file, e.strerror), logger.DEBUG)


def log_helper(message, level=logging.INFO):
    logger.log(level, message)
    return message + "\n"


def process_dir(process_path, release_name=None, process_method=None, force=False, is_priority=None, delete_on=False, failed=False, mode="auto"):
    """
    Scans through the files in process_path and processes whatever media files it finds

    :param process_path: The folder name to look in
    :param release_name: The NZB/Torrent name which resulted in this folder being downloaded
    :param process_method: processing method, copy/move/symlink/link
    :param force: True to process previously processed files
    :param is_priority: whether to replace the file even if it exists at higher quality
    :param delete_on: delete files and folders after they are processed (always happens with move and auto combination)
    :param failed: Boolean for whether or not the download failed
    :param mode: Type of postprocessing auto or manual
    """
    result = ProcessResult()
    try:
        # if they passed us a real dir then assume it's the one we want
        if os.path.isdir(process_path):
            process_path = os.path.realpath(process_path)
            result.output += log_helper("Processing in folder {0}".format(process_path), logger.DEBUG)

        # if the client and SickChill are not on the same machine translate the directory into a network directory
        elif all(
            [settings.TV_DOWNLOAD_DIR, os.path.isdir(settings.TV_DOWNLOAD_DIR), os.path.normpath(process_path) == os.path.normpath(settings.TV_DOWNLOAD_DIR)]
        ):
            process_path = os.path.join(settings.TV_DOWNLOAD_DIR, os.path.abspath(process_path).split(os.path.sep)[-1])
            result.output += log_helper("Trying to use folder: {0} ".format(process_path), logger.DEBUG)

        # if we didn't find a real dir then quit
        if not os.path.isdir(process_path):
            result.output += log_helper(
                "Unable to figure out what folder to process. "
                "If your downloader and SickChill aren't on the same PC "
                "make sure you fill out your TV download dir in the config.",
                logger.DEBUG,
            )
            return result.output

        process_method = process_method or settings.PROCESS_METHOD

        directories_from_rars = set()

        # If we have a release name (probably from nzbToMedia), and it is a rar/video, only process that file
        if release_name and validators.url(release_name):
            result.output += log_helper(_("Processing {0}").format(release_name), logger.INFO)
            generator_to_use = [("", [], [release_name])]
        elif release_name and (helpers.is_media_file(release_name) or helpers.is_rar_file(release_name)):
            result.output += log_helper(_("Processing {0}").format(release_name), logger.INFO)
            generator_to_use = [(process_path, [], [release_name])]
        else:
            result.output += log_helper(_("Processing {0}").format(process_path), logger.INFO)
            generator_to_use = os.walk(process_path, followlinks=settings.PROCESSOR_FOLLOW_SYMLINKS)

        for current_directory, directory_names, filenames in generator_to_use:
            result.result = True

            if current_directory:
                filenames = [f for f in filenames if not is_torrent_or_nzb_file(f)]
                rar_files = [x for x in filenames if helpers.is_rar_file(os.path.join(current_directory, x))]
                if rar_files:
                    extracted_directories = unrar(current_directory, rar_files, force, result)
                    if extracted_directories:
                        for extracted_directory in extracted_directories:
                            if extracted_directory.split(current_directory)[-1] not in directory_names:
                                result.output += log_helper(
                                    _("Adding extracted directory to the list of directories to process: {0}").format(extracted_directory), logger.DEBUG
                                )
                                directories_from_rars.add(extracted_directory)

            if not validate_dir(current_directory, release_name, failed, result):
                continue

            video_files = list(filter(helpers.is_media_file, filenames))
            if video_files:
                process_media(current_directory, video_files, release_name, process_method, force, is_priority, result)
            else:
                result.result = False

            # Delete all file not needed and avoid deleting files if Manual PostProcessing
            if not (process_method == "move" and result.result) or (mode == "manual" and not delete_on):
                continue

            # noinspection PyTypeChecker
            unwanted_files = [x for x in filenames if x in video_files + rar_files]
            if unwanted_files:
                result.output += log_helper(_("Found unwanted files: {0}").format(unwanted_files), logger.DEBUG)

            delete_folder(os.path.join(current_directory, "@eaDir"), False)
            delete_files(current_directory, unwanted_files, result)
            if delete_folder(current_directory, check_empty=not delete_on):
                result.output += log_helper(_("Deleted folder: {0}").format(current_directory), logger.DEBUG)

        # For processing extracted rars, only allow methods 'move' and 'copy'.
        # On different methods fall back to 'move'.
        method_fallback = ("move", process_method)[process_method in ("move", "copy")]

        for directory_from_rar in directories_from_rars:
            process_dir(
                process_path=directory_from_rar,
                release_name=os.path.basename(directory_from_rar),
                process_method=method_fallback,
                force=force,
                is_priority=is_priority,
                delete_on=settings.DELRARCONTENTS or delete_on or method_fallback == "move",
                failed=failed,
                mode=mode,
            )

            # Delete rar file only if the extracted dir was successfully processed
            if mode == "auto" and method_fallback == "move" or mode == "manual" and delete_on:
                this_rar = [rar_file for rar_file in rar_files if os.path.basename(directory_from_rar) == os.path.basename(rar_file).rpartition(".")[0]]
                delete_files(current_directory, this_rar, result)  # Deletes only if result.result == True

            delete_folder(directory_from_rar, settings.DELRARCONTENTS)

        result.output += log_helper((_("Processing Failed"), _("Successfully processed"))[result.aggresult], (logger.WARNING, logger.INFO)[result.aggresult])
        if result.missed_files:
            result.output += log_helper(_("Some items were not processed."))
            for missed_file in result.missed_files:
                result.output += log_helper(missed_file)

        return result.output
    except Exception:
        logger.debug(traceback.format_exc())
        return result.output


def validate_dir(process_path, release_name, failed, result):
    """
    Check if directory is valid for processing

    :param process_path: Directory to check
    :param release_name: Original NZB/Torrent name
    :param failed: Previously failed objects
    :param result: Previous results
    :return: True if dir is valid for processing, False if not
    """

    result.output += log_helper("Processing folder " + process_path, logger.DEBUG)
    upper_name = os.path.basename(process_path).upper()
    if upper_name.startswith("_FAILED_") or upper_name.endswith("_FAILED_") or (os.sep + "_FAILED_") in upper_name or ("_FAILED_" + os.sep) in upper_name:
        result.output += log_helper(_("The directory name indicates it failed to extract."), logger.DEBUG)
        failed = True
    elif (
        upper_name.startswith("_UNDERSIZED_")
        or upper_name.endswith("_UNDERSIZED_")
        or (os.sep + "_UNDERSIZED_") in upper_name
        or ("_UNDERSIZED_" + os.sep) in upper_name
    ):
        result.output += log_helper(_("The directory name indicates that it was previously rejected for being undersized."), logger.DEBUG)
        failed = True
    elif upper_name.startswith("_UNPACK") or upper_name.endswith("_UNPACK") or (os.sep + "_UNPACK") in upper_name or ("_UNPACK" + os.sep) in upper_name:
        result.output += log_helper(_("The directory name indicates that this release is in the process of being unpacked."), logger.DEBUG)
        result.missed_files.append("{0} : Being unpacked".format(process_path))
        return False

    if failed:
        process_failed(process_path, release_name, result)
        result.missed_files.append("{0} : Failed download".format(process_path))
        return False

    if settings.TV_DOWNLOAD_DIR and helpers.real_path(process_path) != helpers.real_path(settings.TV_DOWNLOAD_DIR) and helpers.is_hidden_folder(process_path):
        result.output += log_helper("Ignoring hidden folder: {0}".format(process_path), logger.DEBUG)
        if not process_path.endswith("@eaDir"):
            result.missed_files.append("{0} : Hidden folder".format(process_path))
        return False

    # make sure the dir isn't inside a show dir
    main_db_con = db.DBConnection()
    sql_results = main_db_con.select("SELECT location FROM tv_shows")

    for sqlShow in sql_results:
        if (
            process_path.lower().startswith(os.path.realpath(sqlShow["location"]).lower() + os.sep)
            or process_path.lower() == os.path.realpath(sqlShow["location"]).lower()
        ):

            result.output += log_helper("Cannot process an episode that's already been moved to its show dir, skipping " + process_path, logger.WARNING)
            return False

    for current_directory, directory_names, filenames in os.walk(process_path, topdown=False, followlinks=settings.PROCESSOR_FOLLOW_SYMLINKS):
        sync_files = list(filter(is_sync_file, filenames))
        if sync_files and settings.POSTPONE_IF_SYNC_FILES:
            result.output += log_helper("Found temporary sync files: {0} in path: {1}".format(sync_files, os.path.join(process_path, sync_files[0])))
            result.output += log_helper("Skipping post processing for folder: {0}".format(process_path))
            result.missed_files.append("{0} : Sync files found".format(os.path.join(process_path, sync_files[0])))
            continue

        found_files = list(filter(helpers.is_media_file, filenames))
        if settings.UNPACK == settings.UNPACK_PROCESS_CONTENTS:
            found_files += list(filter(helpers.is_rar_file, filenames))

        if current_directory != settings.TV_DOWNLOAD_DIR and found_files:
            found_files.append(os.path.basename(current_directory))

        for found_file in found_files:
            try:
                NameParser().parse(found_file, cache_result=False)
            except (InvalidNameException, InvalidShowException) as error:
                logger.debug("Could not properly parse a show and episode from [{}]: {}".format(found_file, str(error)))
            else:
                return True

    result.output += log_helper("{0} : No processable items found in folder".format(process_path), logger.DEBUG)
    return False


def unrar(path, rar_files, force, result):
    """
    Extracts RAR files

    :param path: Path to look for files in
    :param rar_files: Names of RAR files
    :param force: process currently processing items
    :param result: Previous results
    :return: List of unpacked file names
    """

    unpacked_dirs = []

    if settings.UNPACK == settings.UNPACK_PROCESS_CONTENTS and rar_files:
        result.output += log_helper("Packed Releases detected: {0}".format(rar_files), logger.DEBUG)
        for archive in rar_files:
            failure = None
            rar_handle = None
            try:
                archive_path = os.path.join(path, archive)
                if already_processed(path, archive, force, result):
                    result.output += log_helper("Archive file already post-processed, extraction skipped: {0}".format(archive_path), logger.DEBUG)
                    continue

                if not helpers.is_rar_file(archive_path):
                    continue

                result.output += log_helper("Checking if archive is valid and contains a video: {0}".format(archive_path), logger.DEBUG)
                rar_handle = RarFile(archive_path)
                if rar_handle.needs_password():
                    # TODO: Add support in settings for a list of passwords to try here with rar_handle.set_password(x)
                    result.output += log_helper("Archive needs a password, skipping: {0}".format(archive_path))
                    continue

                rar_handle.testrar()

                # If there are no video files in the rar, don't extract it
                rar_media_files = list(filter(helpers.is_media_file, rar_handle.namelist()))
                if not rar_media_files:
                    continue

                rar_release_name = archive.rpartition(".")[0]

                # Choose the directory we'll unpack to:
                if settings.UNPACK_DIR and os.path.isdir(settings.UNPACK_DIR):  # verify the unpack dir exists
                    unpack_base_dir = settings.UNPACK_DIR
                else:
                    unpack_base_dir = path
                    if settings.UNPACK_DIR:  # Let user know if we can't unpack there
                        result.output += log_helper("Unpack directory cannot be verified. Using {0}".format(path), logger.DEBUG)

                # Fix up the list for checking if already processed
                rar_media_files = [os.path.join(unpack_base_dir, rar_release_name, rar_media_file) for rar_media_file in rar_media_files]

                skip_rar = False
                for rar_media_file in rar_media_files:
                    check_path, check_file = os.path.split(rar_media_file)
                    if already_processed(check_path, check_file, force, result):
                        result.output += log_helper("Archive file already post-processed, extraction skipped: {0}".format(rar_media_file), logger.DEBUG)
                        skip_rar = True
                        break

                if skip_rar:
                    continue

                rar_extract_path = os.path.join(unpack_base_dir, rar_release_name)
                result.output += log_helper("Unpacking archive: {0}".format(archive), logger.DEBUG)
                rar_handle.extractall(path=rar_extract_path)
                unpacked_dirs.append(rar_extract_path)

            except RarCRCError:
                failure = ("Archive Broken", "Unpacking failed because of a CRC error")
            except RarWrongPassword:
                failure = ("Incorrect RAR Password", "Unpacking failed because of an Incorrect Rar Password")
            except PasswordRequired:
                failure = ("Rar is password protected", "Unpacking failed because it needs a password")
            except RarOpenError:
                failure = (
                    "Rar Open Error, check the parent folder and destination file permissions.",
                    "Unpacking failed with a File Open Error (file permissions?)",
                )
            except RarExecError:
                failure = ("Invalid Rar Archive Usage", "Unpacking Failed with Invalid Rar Archive Usage. Is unrar installed and on the system PATH?")
            except BadRarFile:
                failure = ("Invalid Rar Archive", "Unpacking Failed with an Invalid Rar Archive Error")
            except NeedFirstVolume:
                continue
            except (Exception, Error) as e:
                failure = (str(e), "Unpacking failed")
            finally:
                if rar_handle:
                    del rar_handle

            if failure:
                result.output += log_helper("Failed to extract the archive {0}: {1}".format(archive, failure[0]), logger.WARNING)
                result.missed_files.append("{0} : Unpacking failed: {1}".format(archive, failure[1]))
                result.result = False
                continue

    return unpacked_dirs


def already_processed(process_path, video_file, force, result):
    """
    Check if we already post processed a file

    :param process_path: Directory a file resides in
    :param video_file: File name
    :param force: Force checking when already checking (currently unused)
    :param result: True if file is already postprocessed, False if not
    :return:
    """
    if force:
        return False

    # Avoid processing the same dir again if we use a process method <> move
    main_db_con = db.DBConnection()
    sql_result = main_db_con.select("SELECT release_name FROM tv_episodes WHERE release_name IN (?, ?) LIMIT 1", [process_path, video_file.rpartition(".")[0]])
    if sql_result:
        # result.output += log_helper("You're trying to post process a dir that's already been processed, skipping", logger.DEBUG)
        return True

    # Needed if we have downloaded the same episode @ different quality
    # But we need to make sure we check the history of the episode we're going to PP, and not others
    try:  # if it fails to find any info (because we're doing an unparsable folder (like the TV root dir) it will throw an exception, which we want to ignore
        parse_result = NameParser(process_path, tryIndexers=True).parse(process_path)
    except (
        InvalidNameException,
        InvalidShowException,
    ):  # ignore the exception, because we kind of expected it, but create parse_result anyway so we can perform a check on it.
        parse_result = False

    search_sql = "SELECT tv_episodes.indexerid, history.resource FROM tv_episodes INNER JOIN history ON history.showid=tv_episodes.showid"  # This part is always the same
    search_sql += " WHERE history.season=tv_episodes.season AND history.episode=tv_episodes.episode"

    # If we find a showid, a season number, and one or more episode numbers then we need to use those in the query
    if parse_result and parse_result.show.indexerid and parse_result.episode_numbers and parse_result.season_number:
        search_sql += " AND tv_episodes.showid={0} AND tv_episodes.season={1} AND tv_episodes.episode={2}".format(
            parse_result.show.indexerid, parse_result.season_number, parse_result.episode_numbers[0]
        )

    search_sql += " AND tv_episodes.status IN (" + ",".join([str(x) for x in common.Quality.DOWNLOADED + common.Quality.ARCHIVED]) + ")"
    search_sql += " AND history.resource LIKE ? LIMIT 1"
    sql_result = main_db_con.select(search_sql, ["%" + video_file])
    if sql_result:
        result.output += log_helper("You're trying to post process a video that's already been processed, skipping", logger.DEBUG)
        return True

    return False


def process_media(process_path, video_files, release_name, process_method, force, is_priority, result):
    """
    Postprocess mediafiles

    :param process_path: Path to process in
    :param video_files: Filenames to look for and postprocess
    :param release_name: Name of NZB/Torrent file related
    :param process_method: auto/manual
    :param force: Postprocess currently postprocessing file
    :param is_priority: Boolean, is this a priority download
    :param result: Previous results
    """

    processor = None
    for cur_video_file in video_files:
        cur_video_file_path = os.path.join(process_path, cur_video_file)

        if already_processed(process_path, cur_video_file, force, result):
            result.output += log_helper("Skipping already processed file: {0}".format(cur_video_file), logger.DEBUG)
            continue

        try:
            processor = postProcessor.PostProcessor(cur_video_file_path, release_name, process_method, is_priority)
            result.result = processor.process()
            process_fail_message = ""
        except EpisodePostProcessingFailedException as e:
            result.result = False
            process_fail_message = str(e)

        if processor:
            result.output += processor.log

        if result.result:
            result.output += log_helper("Processing succeeded for {0}".format(cur_video_file_path))
        else:
            result.output += log_helper("Processing failed for {0}: {1}".format(cur_video_file_path, process_fail_message), logger.WARNING)
            result.missed_files.append("{0} : Processing failed: {1}".format(cur_video_file_path, process_fail_message))
            result.aggresult = False


def process_failed(process_path, release_name, result):
    """Process a download that did not complete correctly"""

    if settings.USE_FAILED_DOWNLOADS:
        processor = None

        try:
            processor = failedProcessor.FailedProcessor(process_path, release_name)
            result.result = processor.process()
            process_fail_message = ""
        except FailedPostProcessingFailedException as e:
            result.result = False
            process_fail_message = str(e)

        if processor:
            result.output += processor.log

        if settings.DELETE_FAILED and result.result:
            if delete_folder(process_path, check_empty=False):
                result.output += log_helper("Deleted folder: {0}".format(process_path), logger.DEBUG)

        if result.result:
            result.output += log_helper("Failed Download Processing succeeded: ({0}, {1})".format(release_name, process_path))
        else:
            result.output += log_helper(
                "Failed Download Processing failed: ({0}, {1}): {2}".format(release_name, process_path, process_fail_message), logger.WARNING
            )


def subtitles_enabled(video):
    """
    Parse video filename to a show to check if it has subtitle enabled

    :param video: video filename to be parsed
    """

    try:
        parse_result = NameParser().parse(video, cache_result=True)
    except (InvalidNameException, InvalidShowException):
        logger.warning("Not enough information to parse filename into a valid show. Consider add scene exceptions or improve naming for: {0}".format(video))
        return False

    if parse_result.show.indexerid:
        main_db_con = db.DBConnection()
        sql_results = main_db_con.select("SELECT subtitles FROM tv_shows WHERE indexer_id = ? LIMIT 1", [parse_result.show.indexerid])
        return bool(sql_results[0]["subtitles"]) if sql_results else False
    else:
        logger.warning("Empty indexer ID for: {0}".format(video))
        return False



import datetime
import os
from xml.etree import ElementTree

import sickchill
from sickchill.helper.common import dateFormat, replace_extension
from sickchill.oldbeard import helpers

from ... import logger
from . import mediabrowser


class Mede8erMetadata(mediabrowser.MediaBrowserMetadata):
    """
    Metadata generation class for Mede8er based on the MediaBrowser.

    The following file structure is used:

    show_root/series.xml                    (show metadata)
    show_root/folder.jpg                    (poster)
    show_root/fanart.jpg                    (fanart)
    show_root/Season ##/folder.jpg          (season thumb)
    show_root/Season ##/filename.ext        (*)
    show_root/Season ##/filename.xml        (episode metadata)
    show_root/Season ##/filename.jpg        (episode thumb)
    """

    def __init__(
        self,
        show_metadata=False,
        episode_metadata=False,
        fanart=False,
        poster=False,
        banner=False,
        episode_thumbnails=False,
        season_posters=False,
        season_banners=False,
        season_all_poster=False,
        season_all_banner=False,
    ):

        mediabrowser.MediaBrowserMetadata.__init__(
            self,
            show_metadata,
            episode_metadata,
            fanart,
            poster,
            banner,
            episode_thumbnails,
            season_posters,
            season_banners,
            season_all_poster,
            season_all_banner,
        )

        self.name = "Mede8er"

        self.fanart_name = "fanart.jpg"

        # web-ui metadata template
        # self.eg_show_metadata = "series.xml"
        self.eg_episode_metadata = "Season##\\<i>filename</i>.xml"
        self.eg_fanart = "fanart.jpg"
        # self.eg_poster = "folder.jpg"
        # self.eg_banner = "banner.jpg"
        self.eg_episode_thumbnails = "Season##\\<i>filename</i>.jpg"
        # self.eg_season_posters = "Season##\\folder.jpg"
        # self.eg_season_banners = "Season##\\banner.jpg"
        # self.eg_season_all_poster = "<i>not supported</i>"
        # self.eg_season_all_banner = "<i>not supported</i>"

    def get_episode_file_path(self, ep_obj):
        return replace_extension(ep_obj.location, self._ep_nfo_extension)

    @staticmethod
    def get_episode_thumb_path(ep_obj):
        return replace_extension(ep_obj.location, "jpg")

    def _show_data(self, show_obj):
        """
        Creates an elementTree XML structure for a MediaBrowser-style series.xml
        returns the resulting data object.

        show_obj: a TVShow instance to create the NFO for
        """
        rootNode = ElementTree.Element("details")
        tv_node = ElementTree.SubElement(rootNode, "movie")
        tv_node.attrib["isExtra"] = "false"
        tv_node.attrib["isSet"] = "false"
        tv_node.attrib["isTV"] = "true"

        myShow = sickchill.indexer.series(show_obj)
        if not myShow:
            logger.info("Unable to find show with id {} on {}, skipping it".format(show_obj.indexerid, show_obj.idxr.name))
            return False

        # check for title and id
        if not (getattr(myShow, "seriesName", None) and getattr(myShow, "id", None)):
            logger.info("Incomplete info for show with id {} on {}, skipping it".format(show_obj.indexerid, show_obj.idxr.name))
            return False

        SeriesName = ElementTree.SubElement(tv_node, "title")
        SeriesName.text = myShow.seriesName

        if getattr(myShow, "genre", None):
            Genres = ElementTree.SubElement(tv_node, "genres")
            for genre in myShow.genre:
                if genre and genre.strip():
                    cur_genre = ElementTree.SubElement(Genres, "Genre")
                    cur_genre.text = genre.strip()

        if getattr(myShow, "firstAired", None):
            FirstAired = ElementTree.SubElement(tv_node, "premiered")
            FirstAired.text = myShow.firstAired

        if getattr(myShow, "firstAired", None):
            try:
                year_text = str(datetime.datetime.strptime(myShow.firstAired, dateFormat).year)
                if year_text:
                    year = ElementTree.SubElement(tv_node, "year")
                    year.text = year_text
            except Exception:
                pass

        if getattr(myShow, "overview", None):
            plot = ElementTree.SubElement(tv_node, "plot")
            plot.text = myShow.overview

        if getattr(myShow, "rating", None):
            try:
                rating = int(float(myShow.siteRating) * 10)
            except ValueError:
                rating = 0

            if rating:
                Rating = ElementTree.SubElement(tv_node, "rating")
                Rating.text = str(rating)

        if getattr(myShow, "status", None):
            Status = ElementTree.SubElement(tv_node, "status")
            Status.text = myShow.status

        if getattr(myShow, "contentRating", None):
            mpaa = ElementTree.SubElement(tv_node, "mpaa")
            mpaa.text = myShow.rating

        if getattr(myShow, "imdb_id", None):
            imdb_id = ElementTree.SubElement(tv_node, "id")
            imdb_id.attrib["moviedb"] = "imdb"
            imdb_id.text = myShow.imdbId

        if getattr(myShow, "id", None):
            indexerid = ElementTree.SubElement(tv_node, "indexerid")
            indexerid.text = str(myShow.id)

        if getattr(myShow, "runtime", None):
            Runtime = ElementTree.SubElement(tv_node, "runtime")
            Runtime.text = myShow.runtime

        actors = show_obj.idxr.actors(myShow)
        if actors:
            cast = ElementTree.SubElement(tv_node, "cast")
            for actor in actors:
                if "name" in actor and actor["name"].strip():
                    cur_actor = ElementTree.SubElement(cast, "actor")
                    cur_actor.text = actor["name"].strip()

        helpers.indentXML(rootNode)

        data = ElementTree.ElementTree(rootNode)

        return data

    def _ep_data(self, ep_obj):
        """
        Creates an elementTree XML structure for a MediaBrowser style episode.xml
        and returns the resulting data object.

        show_obj: a TVShow instance to create the NFO for
        """

        eps_to_write = [ep_obj] + ep_obj.relatedEps

        myShow = ep_obj.idxr.series_from_episode(ep_obj)
        if not myShow:
            logger.info("Unable to connect to {} while creating meta files - skipping".format(ep_obj.idxr.name))
            return False

        rootNode = ElementTree.Element("details")
        movie = ElementTree.SubElement(rootNode, "movie")

        movie.attrib["isExtra"] = "false"
        movie.attrib["isSet"] = "false"
        movie.attrib["isTV"] = "true"

        # write an MediaBrowser XML containing info for all matching episodes
        for curEpToWrite in eps_to_write:
            myEp = curEpToWrite.idxr.episode(curEpToWrite)
            if not myEp:
                logger.info(
                    "Metadata writer is unable to find episode {0:d}x{1:d} of {2} on {3}..."
                    "has it been removed? Should I delete from db?".format(curEpToWrite.season, curEpToWrite.episode, curEpToWrite.show.name, ep_obj.idxr.name)
                )
                return None

            if curEpToWrite == ep_obj:
                # root (or single) episode

                if ep_obj.airdate != datetime.date.min and not myEp.get("firstAired"):
                    myEp["firstAired"] = str(ep_obj.airdate)

                if not (myEp.get("episodeName") and myEp.get("firstAired")):
                    return None

                episode = movie

                if curEpToWrite.name:
                    EpisodeName = ElementTree.SubElement(episode, "title")
                    EpisodeName.text = curEpToWrite.name

                SeasonNumber = ElementTree.SubElement(episode, "season")
                SeasonNumber.text = str(curEpToWrite.season)

                EpisodeNumber = ElementTree.SubElement(episode, "episode")
                EpisodeNumber.text = str(curEpToWrite.episode)

                if getattr(myShow, "firstAired", None):
                    try:
                        year_text = str(datetime.datetime.strptime(myShow.firstAired, dateFormat).year)
                        if year_text:
                            year = ElementTree.SubElement(episode, "year")
                            year.text = year_text
                    except Exception:
                        pass

                if getattr(myShow, "overview", None):
                    plot = ElementTree.SubElement(episode, "plot")
                    plot.text = myShow.overview

                if curEpToWrite.description:
                    Overview = ElementTree.SubElement(episode, "episodeplot")
                    Overview.text = curEpToWrite.description

                if getattr(myShow, "contentRating", None):
                    mpaa = ElementTree.SubElement(episode, "mpaa")
                    mpaa.text = myShow.rating

                if not ep_obj.relatedEps and myEp.get("rating"):
                    try:
                        rating = int((float(myEp.siteRating) * 10))
                    except ValueError:
                        rating = 0

                    if rating:
                        Rating = ElementTree.SubElement(episode, "rating")
                        Rating.text = str(rating)

                if myEp.get("directors") and isinstance(myEp["directors"], list):
                    for director in myEp["directors"]:
                        cur_director = ElementTree.SubElement(episode, "director")
                        cur_director.text = director

                if myEp.get("writers") and isinstance(myEp["writers"], list):
                    for writer in myEp["writers"]:
                        cur_writer = ElementTree.SubElement(episode, "credits")
                        cur_writer.text = writer

                if myEp.get("guestStars") and isinstance(myEp["guestStars"], list):
                    cast = ElementTree.SubElement(episode, "cast")
                    for actor in myEp["guestStars"]:
                        cur_actor = ElementTree.SubElement(cast, "actor")
                        cur_actor.text = actor

            else:
                # append data from (if any) related episodes

                if curEpToWrite.name:
                    if not EpisodeName.text:
                        EpisodeName.text = curEpToWrite.name
                    else:
                        EpisodeName.text = EpisodeName.text + ", " + curEpToWrite.name

                if curEpToWrite.description:
                    if not Overview.text:
                        Overview.text = curEpToWrite.description
                    else:
                        Overview.text = Overview.text + "\r" + curEpToWrite.description

        helpers.indentXML(rootNode)
        data = ElementTree.ElementTree(rootNode)

        return data

    def write_show_file(self, show_obj):
        """
        Generates and writes show_obj's metadata under the given path to the
        filename given by get_show_file_path()

        show_obj: TVShow object for which to create the metadata

        path: An absolute or relative path where we should put the file. Note that
                the file name will be the default show_filename.

        Note that this method expects that _show_data will return an ElementTree
        object. If your _show_data returns data in another format yo'll need to
        override this method.
        """

        data = self._show_data(show_obj)

        if not data:
            return False

        nfo_file_path = self.get_show_file_path(show_obj)
        nfo_file_dir = os.path.dirname(nfo_file_path)

        try:
            if not os.path.isdir(nfo_file_dir):
                logger.debug("Metadata dir didn't exist, creating it at " + nfo_file_dir)
                os.makedirs(nfo_file_dir)
                helpers.chmodAsParent(nfo_file_dir)

            logger.debug("Writing show nfo file to " + nfo_file_path)

            nfo_file = open(nfo_file_path, "wb")

            data.write(nfo_file, encoding="utf-8", xml_declaration=True)
            nfo_file.close()
            helpers.chmodAsParent(nfo_file_path)
        except IOError as e:
            logger.error("Unable to write file to " + nfo_file_path + " - are you sure the folder is writable? " + str(e))
            return False

        return True

    def write_ep_file(self, ep_obj):
        """
        Generates and writes ep_obj's metadata under the given path with the
        given filename root. Uses the episode's name with the extension in
        _ep_nfo_extension.

        ep_obj: TVEpisode object for which to create the metadata

        filename_path: The file name to use for this metadata. Note that the extension
                will be automatically added based on _ep_nfo_extension. This should
                include an absolute path.

        Note that this method expects that _ep_data will return an ElementTree
        object. If your _ep_data returns data in another format yo'll need to
        override this method.
        """

        data = self._ep_data(ep_obj)

        if not data:
            return False

        nfo_file_path = self.get_episode_file_path(ep_obj)
        nfo_file_dir = os.path.dirname(nfo_file_path)

        try:
            if not os.path.isdir(nfo_file_dir):
                logger.debug("Metadata dir didn't exist, creating it at " + nfo_file_dir)
                os.makedirs(nfo_file_dir)
                helpers.chmodAsParent(nfo_file_dir)

            logger.debug("Writing episode nfo file to " + nfo_file_path)

            nfo_file = open(nfo_file_path, "wb")

            data.write(nfo_file, encoding="utf-8", xml_declaration=True)
            nfo_file.close()
            helpers.chmodAsParent(nfo_file_path)
        except IOError as e:
            logger.error("Unable to write file to " + nfo_file_path + " - are you sure the folder is writable? " + str(e))
            return False

        return True


# present a standard "interface" from the module
metadata_class = Mede8erMetadata



import re

# Resolutions
resolution = [
    re.compile(r"(?P<vres>4320|2160|1080|720|480|360)(?P<scan>[pi])", re.I),
    re.compile(r"(:?[sph]d)(:?.?tv)?[ \-_\.]*(?P<vres>4320|2160|1080|720|480|360)(:?(?P<scan>[pi])|[^\d]|$)", re.I),
]

# Sources
tv = re.compile(r"([sph]d).?tv|tv(rip|mux)", re.I)
dvd = re.compile(r"(?P<hd>hd)?dvd(?P<rip>rip|mux)?", re.I)
web = re.compile(r"(web(?P<type>rip|mux|hd|uhd|.?dl|\b))", re.I)
bluray = re.compile(r"(blue?-?ray|b[rd](?:rip|mux))", re.I)
sat = re.compile(r"(dsr|satrip)", re.I)
itunes = re.compile(r"itunes(hd|uhd)?", re.I)
netflix = re.compile(r"netflix(hd|uhd)?", re.I)
amazon = re.compile(r"(amzn|amazon)(hd|uhd)?", re.I)

# Codecs
avc = re.compile(r"([xh].?26[45]|AVC)", re.I)
xvid = re.compile(r"(xvid|divx)", re.I)
mpeg = re.compile(r"(mpeg-?2)", re.I)

# anime
anime_sd = re.compile(r"(848x480|480p|360p|xvid)", re.I)
anime_hd = re.compile(r"((1280|960)x720|720p)", re.I)
anime_fullhd = re.compile(r"(1920x1080|1080p)", re.I)
anime_bluray = re.compile(r"(blue?-?ray|b[rd](?:rip|mux)|(?:\b|_)bd(?:\b|_))", re.I)



import json

import requests

from sickchill import logger, settings
from sickchill.oldbeard import common


class Notifier(object):

    ROCKETCHAT_ICON_URL = "https://github.com/SickChill/SickChill/raw/master/sickchill/gui/slick/images/sickchill-sc.png"

    def notify_snatch(self, ep_name):
        if settings.ROCKETCHAT_NOTIFY_SNATCH:
            self._notify_rocketchat(common.notifyStrings[common.NOTIFY_SNATCH] + ": " + ep_name)

    def notify_download(self, ep_name):
        if settings.ROCKETCHAT_NOTIFY_DOWNLOAD:
            self._notify_rocketchat(common.notifyStrings[common.NOTIFY_DOWNLOAD] + ": " + ep_name)

    def notify_subtitle_download(self, ep_name, lang):
        if settings.ROCKETCHAT_NOTIFY_SUBTITLEDOWNLOAD:
            self._notify_rocketchat(common.notifyStrings[common.NOTIFY_SUBTITLE_DOWNLOAD] + " " + ep_name + ": " + lang)

    def notify_git_update(self, new_version="??"):
        if settings.USE_ROCKETCHAT:
            update_text = common.notifyStrings[common.NOTIFY_GIT_UPDATE_TEXT]
            title = common.notifyStrings[common.NOTIFY_GIT_UPDATE]
            self._notify_rocketchat(title + " - " + update_text + new_version)

    def notify_login(self, ipaddress=""):
        if settings.USE_ROCKETCHAT:
            update_text = common.notifyStrings[common.NOTIFY_LOGIN_TEXT]
            title = common.notifyStrings[common.NOTIFY_LOGIN]
            self._notify_rocketchat(title + " - " + update_text.format(ipaddress))

    def test_notify(self):
        return self._notify_rocketchat("This is a test notification from SickChill", force=True)

    def _send_rocketchat(self, message=None):
        rocketchat_webhook = settings.ROCKETCHAT_WEBHOOK
        rocketchat_icon_emoji = settings.ROCKETCHAT_ICON_EMOJI

        logger.info("Sending rocketchat message: " + message)
        logger.info("Sending rocketchat message to url: " + rocketchat_webhook)

        headers = {"Content-Type": "application/json"}
        try:
            r = requests.post(
                rocketchat_webhook,
                data=json.dumps(dict(text=message, attachments=(dict(icon_emoji=rocketchat_icon_emoji, author_icon=self.ROCKETCHAT_ICON_URL)))),
                headers=headers,
            )
            r.raise_for_status()
        except Exception as e:
            logger.exception("Error Sending RocketChat message: " + str(e))
            return False

        return True

    def _notify_rocketchat(self, message="", force=False):
        if not settings.USE_ROCKETCHAT and not force:
            return False

        return self._send_rocketchat(message)



#
# Uses the Synology Download Station API:
# http://download.synology.com/download/Document/DeveloperGuide/Synology_Download_Station_Web_API.pdf

import re
from json import JSONDecodeError
from urllib.parse import unquote, urljoin

from sickchill import logger, settings
from sickchill.oldbeard.clients.generic import GenericClient
from sickchill.providers.GenericProvider import GenericProvider


class Client(GenericClient):
    """
    Class to send torrents/NZBs or links to them to DownloadStation
    """

    def __init__(self, host=None, username=None, password=None):
        """
        Initializes the DownloadStation client
        params: :host: Url to the Download Station API
                :username: Username to use for authentication
                :password: Password to use for authentication
        """
        super().__init__("DownloadStation", host, username, password)

        self.urls = {
            "login": urljoin(self.host, "webapi/auth.cgi"),
            "task": urljoin(self.host, "webapi/entry.cgi"),
            "info": urljoin(self.host, "webapi/DownloadStation/info.cgi"),
        }

        self.url = self.urls["task"]

        generic_errors = {
            100: "Unknown error",
            101: "Invalid parameter",
            102: "The requested API does not exist",
            103: "The requested method does not exist",
            104: "The requested version does not support the functionality",
            105: "The logged in session does not have permission",
            106: "Session timeout",
            107: "Session interrupted by duplicate login",
        }
        self.error_map = {
            "create": {
                400: "File upload failed",
                401: "Max number of tasks reached",
                402: "Destination denied",
                403: "Destination does not exist",
                404: "Invalid task id",
                405: "Invalid task action",
                406: "No default destination",
                407: "Set destination failed",
                408: "File does not exist",
            },
            "login": {
                400: "No such account or incorrect password",
                401: "Account disabled",
                402: "Permission denied",
                403: "2-step verification code required",
                404: "Failed to authenticate 2-step verification code",
            },
        }
        for api_method in self.error_map:
            self.error_map[api_method].update(generic_errors)

        self._task_post_data = {
            "api": "SYNO.DownloadStation2.Task",
            "version": "2",
            "method": "create",
            "session": "DownloadStation",
        }

    def _get_auth(self):
        """
        Authenticates the session with DownloadStation
        """
        if self.session.cookies and self.auth:
            return self.auth

        params = {
            "api": "SYNO.API.Auth",
            "version": 6,
            "method": "login",
            "account": self.username,
            "passwd": self.password,
            "session": "DownloadStation",
            "format": "cookie",
        }

        self.response = self.session.get(self.urls["login"], params=params, verify=False)

        self.auth = self._check_response()
        return self.auth

    def _check_response(self, data=None):
        """
        Checks the response from Download Station, and logs any errors
        params: :data: post data sent in the original request, in case we need to send it with adjusted parameters
                :file: file data being sent with the post request, if any
        """
        try:
            jdata = self.response.json()
        except (ValueError, AttributeError):
            logger.info("Could not convert response to json, check the host:port: {0!r}".format(self.response))
            return False

        if not jdata.get("success"):
            error_code = jdata.get("error", {}).get("code")
            api_method = (data or {}).get("method", "login")
            log_string = self.error_map.get(api_method).get(error_code, None)
            if not log_string:
                logger.info(jdata)
            else:
                logger.info("{0}".format(log_string))

        return jdata.get("success")

    def _get_destination(self, result):
        """
        Determines which destination setting to use depending on result type
        """
        if result.resultType in (GenericProvider.NZB, GenericProvider.NZBDATA):
            destination = settings.SYNOLOGY_DSM_PATH.strip()
        elif result.resultType == GenericProvider.TORRENT:
            destination = settings.TORRENT_PATH.strip()
        else:
            raise AttributeError("Invalid result passed to client when getting destination: resultType {}".format(result.resultType))

        return re.sub(r"^/volume\d/", "", destination).lstrip("/")

    def _set_destination(self, result, destination):
        """
        Determines which destination setting to use depending on result type and sets it to `destination`
        params: :destination: DSM share name
        """
        destination = destination.strip()
        if result.resultType in (GenericProvider.NZB, GenericProvider.NZBDATA):
            settings.SYNOLOGY_DSM_PATH = destination
        elif result.resultType == GenericProvider.TORRENT:
            settings.TORRENT_PATH = destination
        else:
            raise AttributeError("Invalid result passed to client when setting destination")

    def _check_destination(self, result):
        """
        If destination is not set in configuration, grab it from the API
        params: :result: an object subclassing oldbeard.classes.SearchResult
        """
        params = {
            "api": "SYNO.DownloadStation.Info",
            "version": "2",
            "method": "getconfig",
        }

        if not self._get_destination(result):
            try:
                response = self.session.get(self.urls["info"], params=params, verify=False)
                response_json = response.json()
                self._set_destination(result, response_json["data"]["default_destination"])
                logger.info("Destination set to %s", self._get_destination(result))
            except (ValueError, KeyError, JSONDecodeError) as error:
                logger.debug("Get DownloadStation default destination error: {0}".format(error))
                logger.warning("Could not get share destination from DownloadStation for {}, please set it in the settings", result.resultType)
                raise

    def _add_torrent_uri(self, result):
        """
        Sends a magnet, Torrent url or NZB url to DownloadStation
        params: :result: an object subclassing oldbeard.classes.SearchResult
        """
        self._check_destination(result)

        data = self._task_post_data

        if "%3A%2F%2F" in result.url:
            data["url"] = unquote(result.url)
        else:
            data["url"] = result.url

        data["type"] = "url"
        data["create_list"] = "false"
        data["destination"] = self._get_destination(result)

        logger.info('Posted as url with {} destination "{}"'.format(data["api"], data["destination"]))
        self._request(method="post", data=data)
        return self._check_response(data)

    def _add_torrent_file(self, result):
        """
        Sends a Torrent file or NZB file to DownloadStation
        params: :result: an object subclassing oldbeard.classes.SearchResult
        """
        self._check_destination(result)

        data = self._task_post_data

        result_type = result.resultType.replace("data", "")
        files = {result_type: (".".join([result.name, result_type]), result.content)}

        data["type"] = '"file"'
        data["file"] = f'["{result_type}"]'
        data["create_list"] = "false"
        data["destination"] = f'"{self._get_destination(result)}"'

        logger.info("Posted as file with {} destination {}".format(data["api"], data["destination"]))
        self._request(method="post", data=data, files=files)
        return self._check_response(data)

    def sendNZB(self, result):
        """
        Sends an NZB to DownloadStation
        params: :result: an object subclassing oldbeard.classes.SearchResult
        """
        logger.debug(f"Calling {self.name} Client")

        if not (self.auth or self._get_auth()):
            logger.warning("{0}: Authentication Failed".format(self.name))
            return False

        if result.resultType == "nzb":
            return self._add_torrent_uri(result)
        elif result.resultType == "nzbdata":
            return self._add_torrent_file(result)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sickchill',
 'sickchill.adba',
 'sickchill.helper',
 'sickchill.oldbeard',
 'sickchill.oldbeard.clients',
 'sickchill.oldbeard.databases',
 'sickchill.oldbeard.name_parser',
 'sickchill.oldbeard.notifiers',
 'sickchill.oldbeard.providers',
 'sickchill.oldbeard.trakt_api',
 'sickchill.providers',
 'sickchill.providers.metadata',
 'sickchill.providers.nzb',
 'sickchill.providers.subtitle',
 'sickchill.providers.torrent',
 'sickchill.recompiled',
 'sickchill.show',
 'sickchill.show.indexers',
 'sickchill.show.recommendations',
 'sickchill.system',
 'sickchill.tagger',
 'sickchill.update_manager',
 'sickchill.views',
 'sickchill.views.api',
 'sickchill.views.config',
 'sickchill.views.manage']

package_data = \
{'': ['*'],
 'sickchill': ['gui/slick/css/*',
               'gui/slick/css/fonts/Bold/*',
               'gui/slick/css/fonts/BoldItalic/*',
               'gui/slick/css/fonts/ExtraBold/*',
               'gui/slick/css/fonts/ExtraBoldItalic/*',
               'gui/slick/css/fonts/Italic/*',
               'gui/slick/css/fonts/Light/*',
               'gui/slick/css/fonts/LightItalic/*',
               'gui/slick/css/fonts/Regular/*',
               'gui/slick/css/fonts/Semibold/*',
               'gui/slick/css/fonts/SemiboldItalic/*',
               'gui/slick/css/lib/*',
               'gui/slick/css/lib/images/*',
               'gui/slick/css/lib/images/tablesorter/*',
               'gui/slick/fonts/*',
               'gui/slick/images/*',
               'gui/slick/images/bootstrap-formhelpers/*',
               'gui/slick/images/flags/*',
               'gui/slick/images/ico/*',
               'gui/slick/images/indexers/*',
               'gui/slick/images/menu/*',
               'gui/slick/images/network/*',
               'gui/slick/images/network/canvas/*',
               'gui/slick/images/network/s4/*',
               'gui/slick/images/providers/*',
               'gui/slick/images/subtitles/*',
               'gui/slick/images/subtitles/flags/*',
               'gui/slick/images/tablesorter/*',
               'gui/slick/js/*',
               'gui/slick/js/lib/*',
               'gui/slick/views/*',
               'gui/slick/views/layouts/*',
               'gui/slick/views/movies/*',
               'gui/slick/views/movies/layouts/*',
               'locale/*',
               'locale/af_ZA/LC_MESSAGES/*',
               'locale/ar_SA/LC_MESSAGES/*',
               'locale/bg_BG/LC_MESSAGES/*',
               'locale/ca_ES/LC_MESSAGES/*',
               'locale/cs_CZ/LC_MESSAGES/*',
               'locale/da_DK/LC_MESSAGES/*',
               'locale/de_DE/LC_MESSAGES/*',
               'locale/el_GR/LC_MESSAGES/*',
               'locale/en_GB/LC_MESSAGES/*',
               'locale/en_US/LC_MESSAGES/*',
               'locale/es_ES/LC_MESSAGES/*',
               'locale/et_EE/LC_MESSAGES/*',
               'locale/fa_IR/LC_MESSAGES/*',
               'locale/fi_FI/LC_MESSAGES/*',
               'locale/fr_FR/LC_MESSAGES/*',
               'locale/he_IL/LC_MESSAGES/*',
               'locale/hr_HR/LC_MESSAGES/*',
               'locale/hu_HU/LC_MESSAGES/*',
               'locale/id_ID/LC_MESSAGES/*',
               'locale/is_IS/LC_MESSAGES/*',
               'locale/it_IT/LC_MESSAGES/*',
               'locale/ja_JP/LC_MESSAGES/*',
               'locale/ko_KR/LC_MESSAGES/*',
               'locale/lv_LV/LC_MESSAGES/*',
               'locale/nl_NL/LC_MESSAGES/*',
               'locale/no_NO/LC_MESSAGES/*',
               'locale/pl_PL/LC_MESSAGES/*',
               'locale/pt_BR/LC_MESSAGES/*',
               'locale/pt_PT/LC_MESSAGES/*',
               'locale/ro_RO/LC_MESSAGES/*',
               'locale/ru_RU/LC_MESSAGES/*',
               'locale/sk_SK/LC_MESSAGES/*',
               'locale/sl_SI/LC_MESSAGES/*',
               'locale/sv_SE/LC_MESSAGES/*',
               'locale/tr_TR/LC_MESSAGES/*',
               'locale/uk_UA/LC_MESSAGES/*',
               'locale/vi_VN/LC_MESSAGES/*',
               'locale/zh_CN/LC_MESSAGES/*',
               'locale/zh_TW/LC_MESSAGES/*']}

modules = \
['SickChill']
install_requires = \
['CacheControl>=0.12.6,<0.13.0',
 'IMDbPY>=2020.09.25',
 'Js2Py>=0.70,<0.72',
 'Mako>=1.1.4,<2.0.0',
 'PyGithub>=1.54.1,<2.0.0',
 'PyNMA>=1.0,<2.0',
 'PySocks>=1.7.1,<2.0.0',
 'SQLAlchemy>=1.4.3,<2.0.0',
 'Send2Trash>=1.5.0,<2.0.0',
 'Unidecode>=1.2.0,<2.0.0',
 'beautifulsoup4>=4.9.3,<5.0.0',
 'beekeeper>=0.9.2,<0.10.0',
 'bencode.py>=4.0.0,<5.0.0',
 'chardet>=4.0.0,<5.0.0',
 'cloudscraper>=1.2.56,<2.0.0',
 'configobj>=5.0.6,<6.0.0',
 'deluge-client>=1.9.0,<2.0.0',
 'enzyme>=0.4.1,<0.5.0',
 'feedparser>=6.0.2,<7.0.0',
 'gntp>=1.0.3,<2.0.0',
 'guessit>=3.3.1,<4.0.0',
 'html5lib>=1.1,<2.0',
 'httplib2>=0.19.0,<0.20.0',
 'ifaddr>=0.1.7,<0.2.0',
 'imagesize>=1.2.0,<2.0.0',
 'ipaddress>=1.0.23,<2.0.0',
 'jsonrpclib-pelix>=0.4.2,<0.5.0',
 'kodipydent>=0.3.1,<0.4.0',
 'markdown2>=2.4.0,<3.0.0',
 'new-rtorrent-python>=1.0.1-alpha.0,<2.0.0',
 'packaging>=20.9,<21.0',
 'profilehooks>=1.12.0,<2.0.0',
 'putio.py>=8.7.0,<9.0.0',
 'pyOpenSSL>=20.0.1,<21.0.0',
 'pyaes>=1.6.1,<2.0.0',
 'pymediainfo>=5.0.3,<6.0.0',
 'python-dateutil>=2.8.1,<3.0.0',
 'python-slugify>=4.0.1,<6.0.0',
 'python-twitter>=3.5,<4.0',
 'python3-fanart>=2.0.0,<3.0.0',
 'qbittorrent-api>=2021.3.18,<2022.0.0',
 'rarfile>=4.0,<5.0',
 'requests>=2.25.1,<3.0.0',
 'subliminal>=2.1.0,<3.0.0',
 'temp-babelfish>=0.5.6-dev.0,<0.6.0',
 'timeago>=1.0.15,<2.0.0',
 'tmdbsimple>=2.8.0,<3.0.0',
 'tornado>=6.1,<7.0',
 'tus.py>=1.3.4,<2.0.0',
 'tvdbsimple>=1.0.6,<2.0.0',
 'validators>=0.18.2,<0.19.0',
 'win-inet-pton>=1.1.0,<2.0.0']

entry_points = \
{'poetry.application.plugin': ['plugin = '
                               'poetry_date_version_plugin.plugin:VersionPlugin']}

scripts = \
['SickChill.py']

setup_kwargs = {
    'name': 'sickchill',
    'version': '2021.7.14-1',
    'description': 'Automatic Video Library Manager for TV Shows',
    'long_description': 'SickChill [![Build Status](https://travis-ci.org/SickChill/SickChill.svg?branch=master)](https://travis-ci.org/SickChill/SickChill) [![Build status](https://ci.appveyor.com/api/projects/status/s8bb0iqroecnhya2/branch/master?svg=true)](https://ci.appveyor.com/project/miigotu/SickChill/branch/master) [![XO code style](https://img.shields.io/badge/code_style-XO-5ed9c7.svg)](https://github.com/sindresorhus/xo) [![Donate](https://img.shields.io/badge/donations-appreciated-green.svg)](https://github.com/SickChill/SickChill/wiki/Donations)\n====================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================\n\n#### Please do not confuse us with SickRageTV aka Sickrage.ca, which took over the `sickrage/sickrage` repository in Oct 2018\n#### We will never mine bitcoin on your machine, charge for a "service" or to use the software.\n#### We will never store your information, especially site logins to private trackers.\n#### We believe in honesty and loyalty and privacy.\n\nAutomatic Video Library Manager for TV Shows. It watches for new episodes of your favorite shows, and when they are posted it does its magic.\n\n#### Features\n - Kodi/XBMC library updates, poster/banner/fanart downloads, and NFO/TBN generation\n - Configurable automatic episode renaming, sorting, and other processing\n - Easily see what episodes you\'re missing, are airing soon, and more\n - Automatic torrent/nzb searching, downloading, and processing at the qualities you want\n - Largest list of supported torrent and nzb providers, both public and private\n - Can notify Kodi, XBMC, Growl, Trakt, Twitter, and more when new episodes are available\n - Searches TheTVDB.com and AniDB.net for shows, seasons, episodes, and metadata\n - Episode status management allows for mass failing seasons/episodes to force retrying\n - DVD Order numbering for returning the results in DVD order instead of Air-By-Date order\n - Allows you to choose which indexer to have SickChill search its show info from when importing\n - Automatic XEM Scene Numbering/Naming for seasons/episodes\n - Available for any platform, uses a simple HTTP interface\n - Specials and multi-episode torrent/nzb support\n - Automatic subtitles matching and downloading\n - Improved failed download handling\n - DupeKey/DupeScore for NZBGet 12+\n - Real SSL certificate validation\n - Supports Anime shows\n\n#### Installation/Setup\n\nVisit the [Installation & Configuration guides page](https://github.com/SickChill/SickChill/wiki/Installation-&-Configuration-Guides) on our wiki\n\n#### Dependencies\n\nTo run SickChill you will need Python 3.6+\n\n#### More info\n\n* [Wiki](https://github.com/SickChill/SickChill/wiki)\n\n* [FAQ](https://github.com/SickChill/SickChill/wiki/FAQ%27s-and-Fixes)\n\n* [Issue Tracker](https://github.com/SickChill/SickChill/issues)\n\n#### Important notes on switching from other forks\n\nBefore using this with your existing database (`sickbeard.db`) please make a backup copy of it and delete any other database files such as cache.db and failed.db if present.\n\nWe HIGHLY recommend starting out with no database files at all to make this a fresh start but the choice is at your own risk.\n\n#### Supported providers\n\nA full list can be found here: [Link](https://github.com/SickChill/SickChill/wiki/SickChill-Search-Providers)\n\n#### News and Changelog\n[news.md and CHANGES.md have moved to a separate repo, click here](https://github.com/SickChill/SickChill.github.io)\n',
    'author': 'Dustyn Gibson',
    'author_email': 'miigotu@gmail.com',
    'maintainer': 'Dustyn Gibson',
    'maintainer_email': 'miigotu@gmail.com',
    'url': 'https://sickchill.github.io',
    'packages': packages,
    'package_data': package_data,
    'py_modules': modules,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'scripts': scripts,
    'python_requires': '>=3.6,<4',
}


setup(**setup_kwargs)



from sickchill import settings
from sickchill.oldbeard import helpers
from sickchill.oldbeard.notifiers import (  # twilio_notify,
    boxcar2,
    discord,
    emailnotify,
    emby,
    freemobile,
    growl,
    join,
    kodi,
    libnotify,
    matrix,
    nmj,
    nmjv2,
    plex,
    prowl,
    pushalot,
    pushbullet,
    pushover,
    pytivo,
    rocketchat,
    slack,
    synoindex,
    synologynotifier,
    telegram,
    trakt,
    tweet,
)

# home theater / nas
kodi_notifier = kodi.Notifier()
plex_notifier = plex.Notifier()
emby_notifier = emby.Notifier()
nmj_notifier = nmj.Notifier()
nmjv2_notifier = nmjv2.Notifier()
synoindex_notifier = synoindex.Notifier()
synology_notifier = synologynotifier.Notifier()
pytivo_notifier = pytivo.Notifier()

# devices
growl_notifier = growl.Notifier()
prowl_notifier = prowl.Notifier()
libnotify_notifier = libnotify.Notifier()
pushover_notifier = pushover.Notifier()
boxcar2_notifier = boxcar2.Notifier()
pushalot_notifier = pushalot.Notifier()
pushbullet_notifier = pushbullet.Notifier()
freemobile_notifier = freemobile.Notifier()
telegram_notifier = telegram.Notifier()
join_notifier = join.Notifier()
# social
twitter_notifier = tweet.Notifier()
# twilio_notifier = twilio_notify.Notifier()
trakt_notifier = trakt.Notifier()
email_notifier = emailnotify.Notifier()
slack_notifier = slack.Notifier()
rocketchat_notifier = rocketchat.Notifier()
matrix_notifier = matrix.Notifier()
discord_notifier = discord.Notifier()

notifiers = [
    libnotify_notifier,  # Libnotify notifier goes first because it doesn't involve blocking on network activity.
    kodi_notifier,
    plex_notifier,
    nmj_notifier,
    nmjv2_notifier,
    synoindex_notifier,
    synology_notifier,
    pytivo_notifier,
    growl_notifier,
    freemobile_notifier,
    telegram_notifier,
    prowl_notifier,
    pushover_notifier,
    boxcar2_notifier,
    pushalot_notifier,
    pushbullet_notifier,
    twitter_notifier,
    # twilio_notifier,
    trakt_notifier,
    email_notifier,
    slack_notifier,
    rocketchat_notifier,
    matrix_notifier,
    discord_notifier,
    join_notifier,
]


def notify_download(ep_name):
    for n in notifiers:
        n.notify_download(ep_name)


def notify_postprocess(ep_name):
    for n in notifiers:
        n.notify_postprocess(ep_name)


def notify_subtitle_download(ep_name, lang):
    for n in notifiers:
        n.notify_subtitle_download(ep_name, lang)


def notify_snatch(ep_name):
    for n in notifiers:
        n.notify_snatch(ep_name)


def notify_git_update(new_version=""):
    if settings.NOTIFY_ON_UPDATE:
        for n in notifiers:
            if hasattr(n, "notify_git_update"):
                n.notify_git_update(new_version)
            else:
                print(n.__module__)


def notify_login(ipaddress):
    if settings.NOTIFY_ON_LOGIN and not helpers.is_ip_local(ipaddress):
        for n in notifiers:
            if hasattr(n, "notify_login"):
                n.notify_login(ipaddress)
            else:
                print(n.__module__)



import asyncio
import errno
import os
import threading
from socket import error as socket_error
from typing import Any, Dict

from tornado.ioloop import IOLoop
from tornado.platform.asyncio import AnyThreadEventLoopPolicy
from tornado.web import Application, RedirectHandler, StaticFileHandler, url

import sickchill.start
from sickchill import logger, settings
from sickchill.oldbeard.helpers import create_https_certificates, generateApiKey
from sickchill.views import CalendarHandler, LoginHandler, LogoutHandler, MoviesHandler
from sickchill.views.api import ApiHandler, KeyHandler

from .routes import Route

# class Custom404Handler(RequestHandler):
#     startTime = 0.
#
#     def prepare(self):
#         return self.redirect(self.reverse_url('home', ''))
#
#         self.set_status(404)
#         t = PageTemplate(rh=self, filename="404.mako")
#         return self.finish(t.render(title='404', header=_('Oops')))


class SickChillStaticFileHandler(StaticFileHandler):
    @classmethod
    def make_static_url(cls, settings: Dict[str, Any], path: str, include_version: bool = True) -> str:
        url = settings.get("static_url_prefix", "/static/") + path
        if not include_version:
            return url

        custom_settings = settings.copy()
        if "cache/" in path:
            custom_settings["static_path"] = os.path.dirname(sickchill.settings.CACHE_DIR)

        version_hash = cls.get_version(custom_settings, path)
        if not version_hash:
            return url

        return "%s?v=%s" % (url, version_hash)


class SRWebServer(threading.Thread):
    def __init__(self, options=None):
        super().__init__()
        self.daemon = True
        self.alive = True
        self.name = "WEBSERVER"

        self.options = options

        if settings.WEB_HOST and settings.WEB_HOST != "0.0.0.0":
            web_host = settings.WEB_HOST
        else:
            web_host = ("0.0.0.0", "")[settings.WEB_IPV6]

        self.options.update(
            {
                "data_root": os.path.join(settings.PROG_DIR, "gui", settings.GUI_NAME),
                "web_root": settings.WEB_ROOT,
                "host": web_host,
                "enable_https": settings.ENABLE_HTTPS,
                "handle_reverse_proxy": settings.HANDLE_REVERSE_PROXY,
                "log_dir": (None, settings.LOG_DIR)[settings.WEB_LOG],
                # 'username': oldbeard.WEB_USERNAME or '',
                # 'password': oldbeard.WEB_PASSWORD or '',
            }
        )

        self.options.setdefault("port", settings.WEB_PORT or 8081)

        self.server = None

        # video root
        if settings.ROOT_DIRS:
            root_dirs = settings.ROOT_DIRS.split("|")
            self.video_root = root_dirs[int(root_dirs[0]) + 1]
        else:
            self.video_root = None

        # web root
        if self.options["web_root"]:
            settings.WEB_ROOT = self.options["web_root"] = "/" + self.options["web_root"].strip("/")

        # api root
        if not settings.API_KEY:
            settings.API_KEY = generateApiKey()
        self.options["api_root"] = f"{settings.WEB_ROOT}/api/{settings.API_KEY}"

        # tornado setup
        self.enable_https = self.options["enable_https"]

        self.https_cert = None
        if settings.HTTPS_CERT:
            self.https_cert = os.path.realpath(settings.HTTPS_CERT)
            if not os.path.exists(self.https_cert) and not os.path.isabs(self.https_cert):
                self.https_cert = os.path.realpath(os.path.join(settings.DATA_DIR, settings.HTTPS_CERT))

        self.https_key = None
        if settings.HTTPS_KEY:
            self.https_key = os.path.realpath(settings.HTTPS_KEY)
            if not os.path.exists(self.https_key) and not os.path.isabs(self.https_key):
                self.https_key = os.path.realpath(os.path.join(settings.DATA_DIR, settings.HTTPS_KEY))

        if self.enable_https:
            # If either the HTTPS certificate or key do not exist, make some self-signed ones.
            if not (self.https_cert and os.path.exists(self.https_cert) and self.https_key and os.path.exists(self.https_key)):
                if not create_https_certificates(self.https_cert, self.https_key):
                    logger.info("Unable to create CERT/KEY files, disabling HTTPS")
                    settings.ENABLE_HTTPS = self.enable_https = False

            if not (os.path.exists(self.https_cert) and os.path.exists(self.https_key)):
                logger.warning("Disabled HTTPS because of missing CERT and KEY files")
                settings.ENABLE_HTTPS = self.enable_https = False

        asyncio.set_event_loop_policy(AnyThreadEventLoopPolicy())

        # Load the app
        self.app = Application(
            [],
            debug=False,  # enables autoreload, compiled_template_cache, static_hash_cache, serve_traceback - This fixes the 404 page and fixes autoreload for
            #  devs. We could now update without restart possibly if we check DB version hasnt changed!
            autoreload=False,
            gzip=settings.WEB_USE_GZIP,
            cookie_secret=settings.WEB_COOKIE_SECRET,
            login_url=f'{self.options["web_root"]}/login/',
            static_path=self.options["data_root"],
            static_url_prefix=f'{self.options["web_root"]}/',
            static_handler_class=SickChillStaticFileHandler
            # default_handler_class=Custom404Handler
        )

        # Static File Handlers
        self.app.add_handlers(
            ".*$",
            [
                url(
                    rf'{self.options["web_root"]}/(favicon\.ico)',
                    SickChillStaticFileHandler,
                    {"path": os.path.join(self.options["data_root"], "images/ico")},
                    name="favicon",
                ),
                url(
                    rf'{self.options["web_root"]}/images/(.*)',
                    SickChillStaticFileHandler,
                    {"path": os.path.join(self.options["data_root"], "images")},
                    name="images",
                ),
                url(
                    rf'{self.options["web_root"]}/cache/images/(.*)',
                    SickChillStaticFileHandler,
                    {"path": os.path.join(settings.CACHE_DIR, "images")},
                    name="image_cache",
                ),
                url(rf'{self.options["web_root"]}/css/(.*)', SickChillStaticFileHandler, {"path": os.path.join(self.options["data_root"], "css")}, name="css"),
                url(rf'{self.options["web_root"]}/js/(.*)', SickChillStaticFileHandler, {"path": os.path.join(self.options["data_root"], "js")}, name="js"),
                url(
                    rf'{self.options["web_root"]}/fonts/(.*)',
                    SickChillStaticFileHandler,
                    {"path": os.path.join(self.options["data_root"], "fonts")},
                    name="fonts",
                )
                # TODO: WTF is this?
                # url(rf'{self.options["web_root"]}/videos/(.*)', SickChillStaticFileHandler,
                #     {"path": self.video_root}, name='videos')
            ],
        )

        # Main Handlers
        self.app.add_handlers(
            ".*$",
            [
                url(rf'{self.options["api_root"]}(/?.*)', ApiHandler, name="api"),
                url(rf'{self.options["web_root"]}/getkey(/?.*)', KeyHandler, name="get_api_key"),
                url(rf'{self.options["web_root"]}/api/builder', RedirectHandler, {"url": self.options["web_root"] + "/apibuilder/"}, name="apibuilder"),
                url(rf'{self.options["web_root"]}/login(/?)', LoginHandler, name="login"),
                url(rf'{self.options["web_root"]}/logout(/?)', LogoutHandler, name="logout"),
                url(rf'{self.options["web_root"]}/calendar/?', CalendarHandler, name="calendar"),
                url(rf'{self.options["web_root"]}/movies/(?P<route>details)/(?P<slug>.*)/', MoviesHandler, name="movies-details"),
                url(rf'{self.options["web_root"]}/movies/(?P<route>remove)/(?P<pk>.*)/', MoviesHandler, name="movies-remove"),
                url(rf'{self.options["web_root"]}/movies/(?P<route>add)/', MoviesHandler, name="movies-add"),
                url(rf'{self.options["web_root"]}/movies/(?P<route>search)/', MoviesHandler, name="movies-search"),
                url(rf'{self.options["web_root"]}/movies/(?P<route>list)/', MoviesHandler, name="movies-list"),
                url(rf'{self.options["web_root"]}/movies/(.*)', MoviesHandler, name="movies"),
                # routes added by @route decorator
                # Plus naked index with missing web_root prefix
            ]
            + Route.get_routes(self.options["web_root"]),
        )

    def run(self):
        if self.enable_https:
            protocol = "https"
            ssl_options = {"certfile": self.https_cert, "keyfile": self.https_key}
        else:
            protocol = "http"
            ssl_options = None

        logger.info("Starting SickChill on " + protocol + "://" + str(self.options["host"]) + ":" + str(self.options["port"]) + "/")

        try:
            self.server = self.app.listen(
                self.options["port"], self.options["host"], ssl_options=ssl_options, xheaders=settings.HANDLE_REVERSE_PROXY, protocol=protocol
            )
        except socket_error as ex:
            err_msg = ""
            if ex.errno == errno.EADDRINUSE:  # Address/port combination already in use
                if settings.LAUNCH_BROWSER and not self.daemon:
                    sickchill.start.launchBrowser("https" if settings.ENABLE_HTTPS else "http", self.options["port"], settings.WEB_ROOT)
                    logger.info("Launching browser and exiting")
                err_msg = "already in use!"

            logger.info(f"Could not start webserver on port {self.options['port']}: {err_msg or ex}")
            # noinspection PyProtectedMember
            os._exit(1)
        except Exception as ex:
            logger.info(f"Could not start webserver on port {self.options['port']}: {ex}")

            # noinspection PyProtectedMember
            os._exit(1)

        try:
            IOLoop.current().start()
            IOLoop.current().close(True)
        except (IOError, ValueError):
            # Ignore errors like "ValueError: I/O operation on closed kqueue fd". These might be thrown during a reload.
            pass

    def shutdown(self):
        self.alive = False
        IOLoop.current().stop()



import datetime
import functools
import locale

from sickchill import settings

from .network_timezones import sb_timezone

date_presets = (
    "%Y-%m-%d",
    "%a, %Y-%m-%d",
    "%A, %Y-%m-%d",
    "%y-%m-%d",
    "%a, %y-%m-%d",
    "%A, %y-%m-%d",
    "%m/%d/%Y",
    "%a, %m/%d/%Y",
    "%A, %m/%d/%Y",
    "%m/%d/%y",
    "%a, %m/%d/%y",
    "%A, %m/%d/%y",
    "%m-%d-%Y",
    "%a, %m-%d-%Y",
    "%A, %m-%d-%Y",
    "%m-%d-%y",
    "%a, %m-%d-%y",
    "%A, %m-%d-%y",
    "%m.%d.%Y",
    "%a, %m.%d.%Y",
    "%A, %m.%d.%Y",
    "%m.%d.%y",
    "%a, %m.%d.%y",
    "%A, %m.%d.%y",
    "%d-%m-%Y",
    "%a, %d-%m-%Y",
    "%A, %d-%m-%Y",
    "%d-%m-%y",
    "%a, %d-%m-%y",
    "%A, %d-%m-%y",
    "%d/%m/%Y",
    "%a, %d/%m/%Y",
    "%A, %d/%m/%Y",
    "%d/%m/%y",
    "%a, %d/%m/%y",
    "%A, %d/%m/%y",
    "%d.%m.%Y",
    "%a, %d.%m.%Y",
    "%A, %d.%m.%Y",
    "%d.%m.%y",
    "%a, %d.%m.%y",
    "%A, %d.%m.%y",
    "%d. %b %Y",
    "%a, %d. %b %Y",
    "%A, %d. %b %Y",
    "%d. %b %y",
    "%a, %d. %b %y",
    "%A, %d. %b %y",
    "%d. %B %Y",
    "%a, %d. %B %Y",
    "%A, %d. %B %Y",
    "%d. %B %y",
    "%a, %d. %B %y",
    "%A, %d. %B %y",
    "%b %d, %Y",
    "%a, %b %d, %Y",
    "%A, %b %d, %Y",
    "%B %d, %Y",
    "%a, %B %d, %Y",
    "%A, %B %d, %Y",
)

time_presets = ("%I:%M:%S %p", "%H:%M:%S")


# helper class
class static_or_instance(object):
    def __init__(self, func):
        self.func = func

    def __get__(self, instance, owner):
        return functools.partial(self.func, instance)


# subclass datetime.datetime to add function to display custom date and time formats
class sbdatetime(datetime.datetime):
    has_locale = True
    en_US_norm = locale.normalize("en_US.utf-8")

    @static_or_instance
    def convert_to_setting(self, dt=None):
        try:
            if settings.TIMEZONE_DISPLAY == "local":
                return dt.astimezone(sb_timezone) if self is None else self.astimezone(sb_timezone)
            else:
                return dt if self is None else self
        except Exception:
            return dt if self is None else self

    # display Time in SickChill Format
    @static_or_instance
    def sbftime(self, dt=None, show_seconds=False, t_preset=None):
        """
        Display time in SC format
        TODO: Rename this to srftime

        :param dt: datetime object
        :param show_seconds: Boolean, show seconds
        :param t_preset: Preset time format
        :return: time string
        """

        try:
            locale.setlocale(locale.LC_TIME, "")
        except Exception:
            pass

        try:
            if sbdatetime.has_locale:
                locale.setlocale(locale.LC_TIME, "en_US")
        except Exception:
            try:
                if sbdatetime.has_locale:
                    locale.setlocale(locale.LC_TIME, sbdatetime.en_US_norm)
            except Exception:
                sbdatetime.has_locale = False

        strt = ""
        try:
            if self is None:
                if dt is not None:
                    if t_preset is not None:
                        strt = dt.strftime(t_preset)
                    elif show_seconds:
                        strt = dt.strftime(settings.TIME_PRESET_W_SECONDS)
                    else:
                        strt = dt.strftime(settings.TIME_PRESET)
            else:
                if t_preset is not None:
                    strt = self.strftime(t_preset)
                elif show_seconds:
                    strt = self.strftime(settings.TIME_PRESET_W_SECONDS)
                else:
                    strt = self.strftime(settings.TIME_PRESET)
        finally:
            try:
                if sbdatetime.has_locale:
                    locale.setlocale(locale.LC_TIME, "")
            except Exception:
                sbdatetime.has_locale = False

        return strt

    # display Date in SickChill Format
    @static_or_instance
    def sbfdate(self, dt=None, d_preset=None):
        """
        Display date in SC format
        TODO: Rename this to srfdate

        :param dt: datetime object
        :param d_preset: Preset date format
        :return: date string
        """

        try:
            locale.setlocale(locale.LC_TIME, "")
        except Exception:
            pass

        strd = ""
        try:
            if self is None:
                if dt is not None:
                    if d_preset is not None:
                        strd = dt.strftime(d_preset)
                    else:
                        strd = dt.strftime(settings.DATE_PRESET)
            else:
                if d_preset is not None:
                    strd = self.strftime(d_preset)
                else:
                    strd = self.strftime(settings.DATE_PRESET)
        except (ValueError, OSError):
            strd = "UNK"

        finally:

            try:
                locale.setlocale(locale.LC_TIME, "")
            except Exception:
                pass

        return strd

    # display Datetime in SickChill Format
    @static_or_instance
    def sbfdatetime(self, dt=None, show_seconds=False, d_preset=None, t_preset=None):
        """
        Show datetime in SC format
        TODO: Rename this to srfdatetime

        :param dt: datetime object
        :param show_seconds: Boolean, show seconds as well
        :param d_preset: Preset date format
        :param t_preset: Preset time format
        :return: datetime string
        """

        try:
            locale.setlocale(locale.LC_TIME, "")
        except Exception:
            pass

        strd = ""
        try:
            if self is None:
                if dt is not None:
                    if d_preset is not None:
                        strd = dt.strftime(d_preset)
                    else:
                        strd = dt.strftime(settings.DATE_PRESET)
                    try:
                        if sbdatetime.has_locale:
                            locale.setlocale(locale.LC_TIME, "en_US")
                    except Exception:
                        try:
                            if sbdatetime.has_locale:
                                locale.setlocale(locale.LC_TIME, sbdatetime.en_US_norm)
                        except Exception:
                            sbdatetime.has_locale = False
                    if t_preset is not None:
                        strd += ", " + dt.strftime(t_preset)
                    elif show_seconds:
                        strd += ", " + dt.strftime(settings.TIME_PRESET_W_SECONDS)
                    else:
                        strd += ", " + dt.strftime(settings.TIME_PRESET)
            else:
                if d_preset is not None:
                    strd = self.strftime(d_preset)
                else:
                    strd = self.strftime(settings.DATE_PRESET)
                try:
                    if sbdatetime.has_locale:
                        locale.setlocale(locale.LC_TIME, "en_US")
                except Exception:
                    try:
                        if sbdatetime.has_locale:
                            locale.setlocale(locale.LC_TIME, sbdatetime.en_US_norm)
                    except Exception:
                        sbdatetime.has_locale = False
                if t_preset is not None:
                    strd += ", " + self.strftime(t_preset)
                elif show_seconds:
                    strd += ", " + self.strftime(settings.TIME_PRESET_W_SECONDS)
                else:
                    strd += ", " + self.strftime(settings.TIME_PRESET)
        except (ValueError, OSError):
            strd = "UNK"
        finally:
            try:
                if sbdatetime.has_locale:
                    locale.setlocale(locale.LC_TIME, "")
            except Exception:
                sbdatetime.has_locale = False

        return strd



import datetime
import sys

import sickchill
from sickchill.helper.common import dateTimeFormat

from .common import Quality


class SearchResult(object):
    """
    Represents a search result from an indexer.
    """

    def __init__(self, episodes):
        self.provider = None

        # release show object
        self.show = None

        # URL to the NZB/torrent file
        self.url = ""

        # used by some providers to store extra info associated with the result
        self.extraInfo = []

        # list of TVEpisode objects that this result is associated with
        self.episodes = episodes

        # quality of the release
        self.quality = Quality.UNKNOWN

        # release name
        self.name = ""

        # size of the release (-1 = n/a)
        self.size = -1

        # release group
        self.release_group = ""

        # version
        self.version = -1

        # hash
        self.hash = None

        # content
        self.content = None

        self.resultType = ""

        self.priority = 0

    def from_json(self, result_dict):
        self.name = result_dict.get("name")
        self.url = result_dict.get("url")
        self.size = result_dict.get("size")
        self.version = result_dict.get("version")
        self.release_group = result_dict.get("release_group")
        self.quality = int(result_dict.get("quality"))
        self.provider = sickchill.oldbeard.providers.getProviderClass(result_dict.get("provider"))

    @classmethod
    def make_result(cls, result_dict):
        show = sickchill.show.Show.Show._validate_indexer_id(result_dict.get("indexerid"))
        if not show[1]:
            return show[0]

        show = show[1]
        episode_objects = [show.getEpisode(result_dict.get("season"), ep) for ep in result_dict.get("episodes").split("|") if ep]
        result = cls(episode_objects)
        result.from_json(result_dict)
        result.show = show

        return result

    def __str__(self):

        if self.provider is None:
            return "Invalid provider, unable to print self"

        my_string = "{0} @ {1}\n".format(self.provider.name, self.url)
        my_string += "Extra Info:\n"
        for extra in self.extraInfo:
            my_string += " {0}\n".format(extra)

        my_string += "Episodes:\n"
        for ep in self.episodes:
            my_string += " {0}\n".format(ep)

        my_string += "Quality: {0}\n".format(Quality.qualityStrings[self.quality])
        my_string += "Name: {0}\n".format(self.name)
        my_string += "Size: {0}\n".format(self.size)
        my_string += "Release Group: {0}\n".format(self.release_group)

        return my_string


class NZBSearchResult(SearchResult):
    """
    Regular NZB result with an URL to the NZB
    """

    def __init__(self, episodes):
        super().__init__(episodes)
        self.resultType = "nzb"


class NZBDataSearchResult(SearchResult):
    """
    NZB result where the actual NZB XML data is stored in the extraInfo
    """

    def __init__(self, episodes):
        super().__init__(episodes)
        self.resultType = "nzbdata"


class TorrentSearchResult(SearchResult):
    """
    Torrent result with an URL to the torrent
    """

    def __init__(self, episodes):
        super().__init__(episodes)
        self.resultType = "torrent"


class Proper(object):
    def __init__(self, name, url, date, show):
        self.name = name
        self.url = url
        self.date = date
        self.provider = None
        self.quality = Quality.UNKNOWN
        self.release_group = None
        self.version = -1

        self.show = show
        self.indexer = None
        self.indexerid = -1
        self.season = -1
        self.episode = -1
        self.scene_season = -1
        self.scene_episode = -1

    def __str__(self):
        return "{date} {name} {season}x{episode} of {series_id} from {indexer}".format(
            date=self.date, name=self.name, season=self.season, episode=self.episode, series_id=self.indexerid, indexer=sickchill.indexer.name(self.indexer)
        )


class ErrorViewer(object):
    """
    Keeps a static list of UIErrors to be displayed on the UI and allows
    the list to be cleared.
    """

    errors = []

    def __init__(self):
        ErrorViewer.errors = []

    @staticmethod
    def add(error):
        ErrorViewer.errors = [e for e in ErrorViewer.errors if e.message != error.message]
        ErrorViewer.errors.append(error)

    @staticmethod
    def clear():
        ErrorViewer.errors = []

    @staticmethod
    def get():
        return ErrorViewer.errors


class WarningViewer(object):
    """
    Keeps a static list of (warning) UIErrors to be displayed on the UI and allows
    the list to be cleared.
    """

    errors = []

    def __init__(self):
        WarningViewer.errors = []

    @staticmethod
    def add(error):
        WarningViewer.errors = [e for e in WarningViewer.errors if e.message != error.message]
        WarningViewer.errors.append(error)

    @staticmethod
    def clear():
        WarningViewer.errors = []

    @staticmethod
    def get():
        return WarningViewer.errors


class UIError(object):
    """
    Represents an error to be displayed in the web UI.
    """

    def __init__(self, message):
        self.title = sys.exc_info()[-2] or message
        self.message = message
        self.time = datetime.datetime.now().strftime(dateTimeFormat)



import time
from xml.etree import ElementTree

import requests

from sickchill import logger, settings


class Notifier(object):
    @staticmethod
    def notify_snatch(ep_name):
        return False
        # Not implemented: Start the scanner when snatched does not make any sense

    def notify_download(self, ep_name):
        self._notifyNMJ()

    def notify_subtitle_download(self, ep_name, lang):
        self._notifyNMJ()

    @staticmethod
    def notify_git_update(new_version):
        return False
        # Not implemented, no reason to start scanner.

    @staticmethod
    def notify_login(ipaddress=""):
        return False

    def test_notify(self, host):
        return self._sendNMJ(host)

    @staticmethod
    def notify_settings(host, dbloc, instance):
        """
        Retrieves the NMJv2 database location from Popcorn hour

        host: The hostname/IP of the Popcorn Hour server
        dbloc: 'local' for PCH internal hard drive. 'network' for PCH network shares
        instance: Allows for selection of different DB in case of multiple databases

        Returns: True if the settings were retrieved successfully, False otherwise
        """
        try:

            url = "http://{0}:8008/file_operation?arg0=list_user_storage_file&arg1=&arg2={1}&arg3=20&arg4=true&arg5=true&arg6=true&arg7=all&arg8=name_asc&arg9=false&arg10=false".format(
                host, instance
            )
            response = requests.get(url)
            et = ElementTree.fromstring(response.text)
            time.sleep(300.0 / 1000.0)
            for node in et.iter("path"):
                url = "http://{}:8008/metadata_database?arg0=check_database&arg1={}".format(host, node.text.replace("[=]", ""))
                response = requests.get(url)
                xml_db = ElementTree.fromstring(response.text)
                if xml_db.findtext("returnValue") == "0":
                    db_path = xml_db.findtext("database_path")
                    if dbloc == "local" and "localhost" in db_path:
                        settings.NMJv2_HOST = host
                        settings.NMJv2_DATABASE = db_path
                        return True
                    if dbloc == "network" and "://" in db_path:
                        settings.NMJv2_HOST = host
                        settings.NMJv2_DATABASE = db_path
                        return True

        except IOError as e:
            logger.warning("Warning: Couldn't contact popcorn hour on host {0}: {1}".format(host, e))
            return False
        return False

    @staticmethod
    def _sendNMJ(host):
        """
        Sends a NMJ update command to the specified machine

        host: The hostname/IP to send the request to (no port)
        database: The database to send the request to
        mount: The mount URL to use (optional)

        Returns: True if the request succeeded, False otherwise
        """

        # if a host is provided then attempt to open a handle to that URL
        try:
            url = "http://{}:8008/metadata_database?arg0=update_scandir&arg1={}&arg2=&arg3=update_all".format(host, settings.NMJv2_DATABASE)
            logger.debug("NMJ scan update command sent to host: {0}".format(host))
            response1 = requests.get(url)
            time.sleep(300.0 / 1000.0)

            url = "http://{}:8008/metadata_database?arg0=scanner_start&arg1{}&arg2=background&arg3=".format(host, settings.NMJv2_DATABASE)
            logger.debug("Try to mount network drive via url: {0}".format(host))
            response2 = requests.get(url)
        except IOError as e:
            logger.warning("Warning: Couldn't contact popcorn hour on host {0}: {1}".format(host, e))
            return False
        try:
            et = ElementTree.fromstring(response1.text)
            result1 = et.findtext("returnValue")
        except SyntaxError as e:
            logger.exception("Unable to parse XML returned from the Popcorn Hour: update_scandir, {0}".format(e))
            return False
        try:
            et = ElementTree.fromstring(response2.text)
            result2 = et.findtext("returnValue")
        except SyntaxError as e:
            logger.exception("Unable to parse XML returned from the Popcorn Hour: scanner_start, {0}".format(e))
            return False

        # if the result was a number then consider that an error
        error_codes = ["8", "11", "22", "49", "50", "51", "60"]
        error_messages = [
            "Invalid parameter(s)/argument(s)",
            "Invalid database path",
            "Insufficient size",
            "Database write error",
            "Database read error",
            "Open fifo pipe failed",
            "Read only file system",
        ]
        if int(result1) > 0:
            index = error_codes.index(result1)
            logger.exception("Popcorn Hour returned an error: {0}".format((error_messages[index])))
            return False
        else:
            if int(result2) > 0:
                index = error_codes.index(result2)
                logger.exception("Popcorn Hour returned an error: {0}".format((error_messages[index])))
                return False
            else:
                logger.info("NMJv2 started background scan")
                return True

    def _notifyNMJ(self, host=None, force=False):
        """
        Sends a NMJ update command based on the SB config settings

        host: The host to send the command to (optional, defaults to the host in the config)
        database: The database to use (optional, defaults to the database in the config)
        mount: The mount URL (optional, defaults to the mount URL in the config)
        force: If True then the notification will be sent even if NMJ is disabled in the config
        """
        if not settings.USE_NMJv2 and not force:
            logger.debug("Notification for NMJ scan update not enabled, skipping this notification")
            return False

        # fill in omitted parameters
        if not host:
            host = settings.NMJv2_HOST

        logger.debug("Sending scan command for NMJ ")

        return self._sendNMJ(host)



from sickchill import logger, settings
from sickchill.helper.common import try_int
from sickchill.oldbeard.classes import NZBSearchResult
from sickchill.providers.GenericProvider import GenericProvider


class NZBProvider(GenericProvider):
    def __init__(self, name):
        super().__init__(name)

        self.provider_type = GenericProvider.NZB
        self.torznab = False

    @property
    def is_active(self):
        return bool(settings.USE_NZBS) and self.is_enabled

    def _get_result(self, episodes):
        result = NZBSearchResult(episodes)
        if self.torznab or result.url.startswith("magnet"):
            result.resultType = GenericProvider.TORRENT

        return result

    def _get_size(self, item):
        try:
            size = item.get("links")[1].get("length", -1)
        except (AttributeError, IndexError, TypeError):
            size = -1

        if not size:
            logger.debug("The size was not found in the provider response")

        return try_int(size, -1)

    def _get_storage_dir(self):
        return settings.NZB_DIR



from random import shuffle


class AniDBMapper:
    blacklist = ("unused", "retired", "reserved")

    def getAnimeBitsA(self, amask):
        map = self.getAnimeMapA()
        return self._getBitChain(map, amask)

    def getAnimeCodesA(self, aBitChain):
        amap = self.getAnimeMapA()
        return self._getCodes(amap, aBitChain)

    def getFileBitsF(self, fmask):
        fmap = self.getFileMapF()
        return self._getBitChain(fmap, fmask)

    def getFileCodesF(self, bitChainF):
        fmap = self.getFileMapF()
        return self._getCodes(fmap, bitChainF)

    def getFileBitsA(self, amask):
        amap = self.getFileMapA()
        return self._getBitChain(amap, amask)

    def getFileCodesA(self, bitChainA):
        amap = self.getFileMapA()
        return self._getCodes(amap, bitChainA)

    def _getBitChain(self, map, wanted):
        """Return an hex string with the correct bit set corresponding to the wanted fields in the map"""
        bit = 0
        for index, field in enumerate(map):
            if field in wanted and not field in self.blacklist:
                bit = bit ^ (1 << len(map) - index - 1)

        bit = str(hex(bit)).lstrip("0x").rstrip("L")
        bit = "".join(["0" for unused in range(len(map) // 4 - len(bit))]) + bit
        return bit

    @staticmethod
    def _getCodes(map, bitChain):
        """Returns a list with the corresponding fields as set in the bitChain (hex string)"""
        codeList = []
        bitChain = int(bitChain, 16)
        mapLength = len(map)
        for i in reversed(list(range(mapLength))):
            if bitChain & (2 ** i):
                codeList.append(map[mapLength - i - 1])
        return codeList

    @staticmethod
    def getAnimeMapA():
        # each line is one byte
        # only chnage this if the api changes
        map = [
            "aid",
            "unused",
            "year",
            "type",
            "related_aid_list",
            "related_aid_type",
            "category_list",
            "category_weight_list",
            "romaji_name",
            "kanji_name",
            "english_name",
            "other_name",
            "short_name_list",
            "synonym_list",
            "retired",
            "retired",
            "episodes",
            "highest_episode_number",
            "special_ep_count",
            "air_date",
            "end_date",
            "url",
            "picname",
            "category_id_list",
            "rating",
            "vote_count",
            "temp_rating",
            "temp_vote_count",
            "average_review_rating",
            "review_count",
            "award_list",
            "is_18_restricted",
            "anime_planet_id",
            "ANN_id",
            "allcinema_id",
            "AnimeNfo_id",
            "unused",
            "unused",
            "unused",
            "date_record_updated",
            "character_id_list",
            "creator_id_list",
            "main_creator_id_list",
            "main_creator_name_list",
            "unused",
            "unused",
            "unused",
            "unused",
            "specials_count",
            "credits_count",
            "other_count",
            "trailer_count",
            "parody_count",
            "unused",
            "unused",
            "unused",
        ]
        return map

    @staticmethod
    def getFileMapF():
        # each line is one byte
        # only chnage this if the api changes
        map = [
            "unused",
            "aid",
            "eid",
            "gid",
            "mylist_id",
            "list_other_episodes",
            "IsDeprecated",
            "state",
            "size",
            "ed2k",
            "md5",
            "sha1",
            "crc32",
            "unused",
            "unused",
            "reserved",
            "quality",
            "source",
            "audio_codec_list",
            "audio_bitrate_list",
            "video_codec",
            "video_bitrate",
            "video_resolution",
            "file_type_extension",
            "dub_language",
            "sub_language",
            "length_in_seconds",
            "description",
            "aired_date",
            "unused",
            "unused",
            "anidb_file_name",
            "mylist_state",
            "mylist_filestate",
            "mylist_viewed",
            "mylist_viewdate",
            "mylist_storage",
            "mylist_source",
            "mylist_other",
            "unused",
        ]
        return map

    @staticmethod
    def getFileMapA():
        # each line is one byte
        # only chnage this if the api changes
        map = [
            "anime_total_episodes",
            "highest_episode_number",
            "year",
            "type",
            "related_aid_list",
            "related_aid_type",
            "category_list",
            "reserved",
            "romaji_name",
            "kanji_name",
            "english_name",
            "other_name",
            "short_name_list",
            "synonym_list",
            "retired",
            "retired",
            "epno",
            "ep_name",
            "ep_romaji_name",
            "ep_kanji_name",
            "episode_rating",
            "episode_vote_count",
            "unused",
            "unused",
            "group_name",
            "group_short_name",
            "unused",
            "unused",
            "unused",
            "unused",
            "unused",
            "date_aid_record_updated",
        ]
        return map

    def checkMapping(self, verbos=False):

        print("------")
        print("File F: " + str(self.checkMapFileF(verbos)))
        print("------")
        print("File A: " + str(self.checkMapFileA(verbos)))

    def checkMapFileF(self, verbos=False):
        return self._checkMapGeneral(self.getFileMapF, self.getFileBitsF, self.getFileCodesF, verbos=verbos)

    def checkMapFileA(self, verbos=False):
        return self._checkMapGeneral(self.getFileMapA, self.getFileBitsA, self.getFileCodesA, verbos=verbos)

    def _checkMapGeneral(self, getGeneralMap, getBits, getCodes, verbos=False):
        map = getGeneralMap()
        shuffle(map)
        mask = [elem for elem in map if elem not in self.blacklist][:5]
        bits = getBits(mask)
        mask_re = getCodes(bits)
        bits_re = getBits(mask_re)
        if verbos:
            print(mask)
            print(mask_re)
            print(bits)
            print(bits_re)
            print("bits are:" + str((bits_re == bits)))
            print("map is :" + str((sorted(mask_re) == sorted(mask))))
        return (bits_re == bits) and sorted(mask_re) == sorted(mask)



import re
import telnetlib
import urllib.parse
import urllib.request
from xml.etree import ElementTree

from sickchill import logger, settings


class Notifier(object):
    @staticmethod
    def notify_settings(host):
        """
        Retrieves the settings from a NMJ/Popcorn hour

        host: The hostname/IP of the Popcorn Hour server

        Returns: True if the settings were retrieved successfully, False otherwise
        """

        # establish a terminal session to the PC
        try:
            terminal = telnetlib.Telnet(host)
        except Exception:
            logger.warning("Warning: unable to get a telnet session to {0}".format(host))
            return False

        # tell the terminal to output the necessary info to the screen so we can search it later
        logger.debug("Connected to {0} via telnet".format(host))
        terminal.read_until("sh-3.00# ")
        terminal.write("cat /tmp/source\n")
        terminal.write("cat /tmp/netshare\n")
        terminal.write("exit\n")
        tnoutput = terminal.read_all()

        match = re.search(r"(.+\.db)\r\n?(.+)(?=sh-3.00# cat /tmp/netshare)", tnoutput)

        # if we found the database in the terminal output then save that database to the config
        if match:
            database = match.group(1)
            device = match.group(2)
            logger.debug("Found NMJ database {0} on device {1}".format(database, device))
            settings.NMJ_DATABASE = database
        else:
            logger.warning("Could not get current NMJ database on {0}, NMJ is probably not running!".format(host))
            return False

        # if the device is a remote host then try to parse the mounting URL and save it to the config
        if device.startswith("NETWORK_SHARE/"):
            match = re.search(".*(?=\r\n?{0})".format((re.escape(device[14:]))), tnoutput)

            if match:
                mount = match.group().replace("127.0.0.1", host)
                logger.debug("Found mounting url on the Popcorn Hour in configuration: {0}".format(mount))
                settings.NMJ_MOUNT = mount
            else:
                logger.warning("Detected a network share on the Popcorn Hour, but could not get the mounting url")
                return False

        return True

    @staticmethod
    def notify_snatch(ep_name):
        return False
        # Not implemented: Start the scanner when snatched does not make any sense

    def notify_download(self, ep_name):
        if settings.USE_NMJ:
            self._notifyNMJ()

    def notify_subtitle_download(self, ep_name, lang):
        if settings.USE_NMJ:
            self._notifyNMJ()

    @staticmethod
    def notify_git_update(new_version):
        return False
        # Not implemented, no reason to start scanner.

    @staticmethod
    def notify_login(ipaddress=""):
        return False

    def test_notify(self, host, database, mount):
        return self._sendNMJ(host, database, mount)

    @staticmethod
    def _sendNMJ(host, database, mount=None):
        """
        Sends a NMJ update command to the specified machine

        host: The hostname/IP to send the request to (no port)
        database: The database to send the request to
        mount: The mount URL to use (optional)

        Returns: True if the request succeeded, False otherwise
        """

        # if a mount URL is provided then attempt to open a handle to that URL
        if mount:
            try:
                req = urllib.request.Request(mount)
                logger.debug("Try to mount network drive via url: {0}".format(mount))
                handle = urllib.request.urlopen(req)
            except IOError as e:
                if hasattr(e, "reason"):
                    logger.warning("NMJ: Could not contact Popcorn Hour on host {0}: {1}".format(host, e.reason))
                elif hasattr(e, "code"):
                    logger.warning("NMJ: Problem with Popcorn Hour on host {0}: {1}".format(host, e.code))
                return False
            except Exception as e:
                logger.exception("NMJ: Unknown exception: " + str(e))
                return False

        # build up the request URL and parameters
        UPDATE_URL = "http://%(host)s:8008/metadata_database?%(params)s"
        params = {"arg0": "scanner_start", "arg1": database, "arg2": "background", "arg3": ""}
        params = urllib.parse.urlencode(params)
        updateUrl = UPDATE_URL % {"host": host, "params": params}

        # send the request to the server
        try:
            req = urllib.request.Request(updateUrl)
            logger.debug("Sending NMJ scan update command via url: {0}".format(updateUrl))
            handle = urllib.request.urlopen(req)
            response = handle.read()
        except IOError as e:
            if hasattr(e, "reason"):
                logger.warning("NMJ: Could not contact Popcorn Hour on host {0}: {1}".format(host, e.reason))
            elif hasattr(e, "code"):
                logger.warning("NMJ: Problem with Popcorn Hour on host {0}: {1}".format(host, e.code))
            return False
        except Exception as e:
            logger.exception("NMJ: Unknown exception: " + str(e))
            return False

        # try to parse the resulting XML
        try:
            et = ElementTree.fromstring(response)
            result = et.findtext("returnValue")
        except SyntaxError as e:
            logger.exception("Unable to parse XML returned from the Popcorn Hour: {0}".format(e))
            return False

        # if the result was a number then consider that an error
        if int(result) > 0:
            logger.exception("Popcorn Hour returned an error code: {0}".format(result))
            return False
        else:
            logger.info("NMJ started background scan")
            return True

    def _notifyNMJ(self, host=None, database=None, mount=None, force=False):
        """
        Sends a NMJ update command based on the SB config settings

        host: The host to send the command to (optional, defaults to the host in the config)
        database: The database to use (optional, defaults to the database in the config)
        mount: The mount URL (optional, defaults to the mount URL in the config)
        force: If True then the notification will be sent even if NMJ is disabled in the config
        """
        if not settings.USE_NMJ and not force:
            logger.debug("Notification for NMJ scan update not enabled, skipping this notification")
            return False

        # fill in omitted parameters
        if not host:
            host = settings.NMJ_HOST
        if not database:
            database = settings.NMJ_DATABASE
        if not mount:
            mount = settings.NMJ_MOUNT

        logger.debug("Sending scan command for NMJ ")

        return self._sendNMJ(host, database, mount)



import os

from tornado.web import addslash

import sickchill.start
from sickchill import logger, settings
from sickchill.helper import try_int
from sickchill.oldbeard import config, filters, helpers, ui
from sickchill.views.common import PageTemplate
from sickchill.views.routes import Route

from . import Config


@Route("/config/search(/?.*)", name="config:search")
class ConfigSearch(Config):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @addslash
    def index(self, *args_, **kwargs_):
        t = PageTemplate(rh=self, filename="config_search.mako")

        return t.render(
            submenu=self.ConfigMenu(), title=_("Config - Episode Search"), header=_("Search Settings"), topmenu="config", controller="config", action="search"
        )

    def saveSearch(
        self,
        use_nzbs=None,
        use_torrents=None,
        nzb_dir=None,
        sab_username=None,
        sab_password=None,
        sab_apikey=None,
        sab_category=None,
        sab_category_anime=None,
        sab_category_backlog=None,
        sab_category_anime_backlog=None,
        sab_host=None,
        nzbget_username=None,
        nzbget_password=None,
        nzbget_category=None,
        nzbget_category_backlog=None,
        nzbget_category_anime=None,
        nzbget_category_anime_backlog=None,
        nzbget_priority=None,
        nzbget_host=None,
        nzbget_use_https=None,
        backlog_days=None,
        backlog_frequency=None,
        dailysearch_frequency=None,
        nzb_method=None,
        torrent_method=None,
        usenet_retention=None,
        download_propers=None,
        check_propers_interval=None,
        allow_high_priority=None,
        sab_forced=None,
        randomize_providers=None,
        use_failed_downloads=None,
        delete_failed=None,
        backlog_missing_only=None,
        torrent_dir=None,
        torrent_username=None,
        torrent_password=None,
        torrent_host=None,
        torrent_label=None,
        torrent_label_anime=None,
        torrent_path=None,
        torrent_path_incomplete=None,
        torrent_verify_cert=None,
        torrent_seed_time=None,
        torrent_paused=None,
        torrent_high_bandwidth=None,
        torrent_rpcurl=None,
        torrent_auth_type=None,
        ignore_words=None,
        trackers_list=None,
        require_words=None,
        ignored_subs_list=None,
        syno_dsm_host=None,
        syno_dsm_user=None,
        syno_dsm_pass=None,
        syno_dsm_path=None,
        quality_allow_hevc=False,
        prefer_words=None,
    ):

        results = []

        if not config.change_nzb_dir(nzb_dir):
            results += ["Unable to create directory " + os.path.normpath(nzb_dir) + ", dir not changed."]

        if not config.change_torrent_dir(torrent_dir):
            results += ["Unable to create directory " + os.path.normpath(torrent_dir) + ", dir not changed."]

        config.change_daily_search_frequency(dailysearch_frequency)

        config.change_backlog_frequency(backlog_frequency)
        settings.BACKLOG_DAYS = try_int(backlog_days, 7)

        settings.USE_NZBS = config.checkbox_to_value(use_nzbs)
        settings.USE_TORRENTS = config.checkbox_to_value(use_torrents)

        settings.NZB_METHOD = nzb_method
        settings.TORRENT_METHOD = torrent_method
        settings.USENET_RETENTION = try_int(usenet_retention, 500)

        settings.IGNORE_WORDS = ignore_words if ignore_words else ""
        settings.TRACKERS_LIST = trackers_list if trackers_list else ""
        settings.REQUIRE_WORDS = require_words if require_words else ""
        settings.PREFER_WORDS = prefer_words if prefer_words else ""
        settings.IGNORED_SUBS_LIST = ignored_subs_list if ignored_subs_list else ""

        settings.RANDOMIZE_PROVIDERS = config.checkbox_to_value(randomize_providers)

        config.change_download_propers(download_propers)

        settings.CHECK_PROPERS_INTERVAL = check_propers_interval

        settings.ALLOW_HIGH_PRIORITY = config.checkbox_to_value(allow_high_priority)
        settings.QUALITY_ALLOW_HEVC = config.checkbox_to_value(quality_allow_hevc)

        settings.USE_FAILED_DOWNLOADS = config.checkbox_to_value(use_failed_downloads)
        settings.DELETE_FAILED = config.checkbox_to_value(delete_failed)

        settings.BACKLOG_MISSING_ONLY = config.checkbox_to_value(backlog_missing_only)

        settings.SAB_USERNAME = sab_username
        settings.SAB_PASSWORD = filters.unhide(settings.SAB_PASSWORD, sab_password)
        settings.SAB_APIKEY = filters.unhide(settings.SAB_APIKEY, sab_apikey.strip())
        settings.SAB_CATEGORY = sab_category
        settings.SAB_CATEGORY_BACKLOG = sab_category_backlog
        settings.SAB_CATEGORY_ANIME = sab_category_anime
        settings.SAB_CATEGORY_ANIME_BACKLOG = sab_category_anime_backlog
        settings.SAB_HOST = config.clean_url(sab_host)
        settings.SAB_FORCED = config.checkbox_to_value(sab_forced)

        settings.NZBGET_USERNAME = nzbget_username
        settings.NZBGET_PASSWORD = filters.unhide(settings.NZBGET_PASSWORD, nzbget_password)
        settings.NZBGET_CATEGORY = nzbget_category
        settings.NZBGET_CATEGORY_BACKLOG = nzbget_category_backlog
        settings.NZBGET_CATEGORY_ANIME = nzbget_category_anime
        settings.NZBGET_CATEGORY_ANIME_BACKLOG = nzbget_category_anime_backlog
        settings.NZBGET_HOST = config.clean_host(nzbget_host)
        settings.NZBGET_USE_HTTPS = config.checkbox_to_value(nzbget_use_https)
        settings.NZBGET_PRIORITY = try_int(nzbget_priority, 100)

        settings.TORRENT_USERNAME = torrent_username
        settings.TORRENT_PASSWORD = filters.unhide(settings.TORRENT_PASSWORD, torrent_password)
        settings.TORRENT_LABEL = torrent_label
        settings.TORRENT_LABEL_ANIME = torrent_label_anime
        settings.TORRENT_VERIFY_CERT = config.checkbox_to_value(torrent_verify_cert)

        settings.TORRENT_PATH = torrent_path.rstrip("/\\")
        settings.TORRENT_PATH_INCOMPLETE = torrent_path_incomplete.rstrip("/\\")

        settings.TORRENT_SEED_TIME = torrent_seed_time
        settings.TORRENT_PAUSED = config.checkbox_to_value(torrent_paused)
        settings.TORRENT_HIGH_BANDWIDTH = config.checkbox_to_value(torrent_high_bandwidth)
        settings.TORRENT_HOST = config.clean_url(torrent_host)
        settings.TORRENT_RPCURL = torrent_rpcurl
        settings.TORRENT_AUTH_TYPE = torrent_auth_type

        settings.SYNOLOGY_DSM_HOST = config.clean_url(syno_dsm_host)
        settings.SYNOLOGY_DSM_USERNAME = syno_dsm_user
        settings.SYNOLOGY_DSM_PASSWORD = filters.unhide(settings.SYNOLOGY_DSM_PASSWORD, syno_dsm_pass)
        settings.SYNOLOGY_DSM_PATH = syno_dsm_path.rstrip("/\\")

        # This is a PITA, but lets merge the settings if they only set DSM up in one section to save them some time
        if settings.TORRENT_METHOD == "download_station":
            if not settings.SYNOLOGY_DSM_HOST:
                settings.SYNOLOGY_DSM_HOST = settings.TORRENT_HOST
            if not settings.SYNOLOGY_DSM_USERNAME:
                settings.SYNOLOGY_DSM_USERNAME = settings.TORRENT_USERNAME
            if not settings.SYNOLOGY_DSM_PASSWORD:
                settings.SYNOLOGY_DSM_PASSWORD = settings.TORRENT_PASSWORD
            if not settings.SYNOLOGY_DSM_PATH:
                settings.SYNOLOGY_DSM_PATH = settings.TORRENT_PATH

        if settings.NZB_METHOD == "download_station":
            if not settings.TORRENT_HOST:
                settings.TORRENT_HOST = settings.SYNOLOGY_DSM_HOST
            if not settings.TORRENT_USERNAME:
                settings.TORRENT_USERNAME = settings.SYNOLOGY_DSM_USERNAME
            if not settings.TORRENT_PASSWORD:
                settings.TORRENT_PASSWORD = settings.SYNOLOGY_DSM_PASSWORD
            if not settings.TORRENT_PATH:
                settings.TORRENT_PATH = settings.SYNOLOGY_DSM_PATH

        helpers.manage_torrents_url(reset=True)

        sickchill.start.save_config()

        if results:
            for x in results:
                logger.exception(x)
            ui.notifications.error(_("Error(s) Saving Configuration"), "<br>\n".join(results))
        else:
            ui.notifications.message(_("Configuration Saved"), os.path.join(settings.CONFIG_FILE))

        return self.redirect("/config/search/")
